import io
import sys
from typing import TYPE_CHECKING

from PyQt6.QtCore import Qt, QTimer, pyqtSignal
from PyQt6.QtWidgets import QLineEdit, QSizePolicy

from feeluown.i18n import t
from feeluown.library.text2song import create_dummy_brief_song
from feeluown.fuoexec import fuoexec
from feeluown.utils.aio import run_afn

if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp

_KeyPrefix = "search_"  # local storage key prefix
KeySourceIn = _KeyPrefix + "source_in"
KeyType = _KeyPrefix + "type"


class MagicBox(QLineEdit):
    """Read user input, parse and execute

    ref: https://wiki.qt.io/Technical_FAQ#How_can_I_create_a_one-line_QTextEdit?
    """

    # this filter signal is designed for table (songs_table & albums_table)
    filter_text_changed = pyqtSignal(str)

    def __init__(self, app: "GuiApp", parent=None):
        super().__init__(parent)

        self._app = app
        self.setPlaceholderText(t("search-box-placeholder"))
        self.setToolTip(t("search-box-tooltip"))
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        self.setFixedHeight(32)
        self.setFrame(False)
        self.setAttribute(Qt.WidgetAttribute.WA_MacShowFocusRect, False)
        self.setTextMargins(5, 0, 0, 0)

        self._timer = QTimer(self)
        self._cmd_text = None
        self._mode = "cmd"  # 详见 _set_mode 函数
        self._timer.timeout.connect(self.__on_timeout)

        self.textChanged.connect(self.__on_text_edited)
        # self.textEdited.connect(self.__on_text_edited)
        self.returnPressed.connect(self.__on_return_pressed)

    def _set_mode(self, mode):
        """Modify current mode

        Currently, there are two main modes: cmd mode is the normal mode;
        msg mode is for displaying message notifications.

        When in msg mode, all signals will be blocked.
        """
        if mode == "cmd":
            self.setReadOnly(False)
            self._timer.stop()
            self.setText(self._cmd_text or "")
            # Note: only disable blockSignals after all operations are completed
            # Then modify the current mode
            self.blockSignals(False)
            self._mode = mode
        elif mode == "msg":
            self.blockSignals(True)
            if self._mode == "cmd":
                self.setReadOnly(True)
                self._cmd_text = self.text()
                self._mode = mode

    def _exec_code(self, code):
        """Execute code and redirect the code’s stdout/stderr."""
        output = io.StringIO()
        sys.stderr = output
        sys.stdout = output
        try:
            obj = compile(code, "<string>", "single")
            fuoexec(obj)
        except Exception as e:
            print(str(e))
        finally:
            sys.stderr = sys.__stderr__
            sys.stdout = sys.__stdout__

        text = output.getvalue()
        self._set_mode("msg")
        self.setText(text or "No output.")
        self._timer.start(1000)
        if text:
            self._app.show_msg(text)

    def __on_text_edited(self):
        text = self.text()
        if self._mode == "cmd":
            self._cmd_text = text
        # filter browser content if prefix starts with `#`
        # TODO: cancel unneeded filter
        if text.startswith("#"):
            self.filter_text_changed.emit(text[1:].strip())
        else:
            self.filter_text_changed.emit("")

    def __on_return_pressed(self):
        text = self.text()
        if text.startswith(">>> "):
            self._exec_code(text[4:])
        elif text.startswith("---") or text.startswith("==="):
            if self._app.ui.ai_chat_overlay is not None:
                body = text[4:] if len(text) > 4 else ""
                if body:
                    run_afn(self._app.ui.ai_chat_overlay.body.exec_user_query, body)
                self._app.ui.ai_chat_overlay.show()
            else:
                self._app.show_msg(t("search-box-ai-chat-unavailable"))
        elif (
            text.startswith("--> ")
            or text.startswith("==> ")
            or text.startswith("--》")
            or text.startswith("==》")
        ):
            body = text[4:]
            if not body:
                return
            delimiters = ("|", "-")
            title = artists_name = ""
            for delimiter in delimiters:
                parts = body.split(delimiter)
                if len(parts) == 2:
                    title, artists_name = parts
                    break
            if title and artists_name:
                song = create_dummy_brief_song(title.strip(), artists_name.strip())
                self._app.playlist.play_model(song)
                self._app.show_msg(t("search-box-play-track", song=song))
            else:
                self._app.show_msg(t("search-box-play-track-ill-formed"))
        else:
            local_storage = self._app.browser.local_storage
            type_ = local_storage.get(KeyType)
            source_in = local_storage.get(KeySourceIn)
            query = {"q": text}
            if type_ is not None:
                query["type"] = type_
            if source_in is not None:
                query["source_in"] = source_in
            self._app.browser.goto(page="/search", query=query)

    def __on_timeout(self):
        self._set_mode("cmd")

    def focusInEvent(self, e):
        super().focusInEvent(e)
        self._set_mode("cmd")



from contextlib import contextmanager
from unittest.mock import MagicMock

from feeluown.app import create_config


@contextmanager
def mock_app():
    app = MagicMock()
    app.config = create_config()
    yield app



import os
import sys
import ctypes
from PyInstaller.utils.hooks import collect_data_files, collect_entry_point


assets_files = [
    '**/*.png',
    '**/*.qss',
    '**/*.svg',
    '**/*.ico',
    '**/*.icns',
    '**/*.colors',
    '**/*.mo',
    '**/*.ftl',
]
datas, hiddenimports = collect_entry_point('fuo.plugins_v1')

if 'fuo_ytmusic' in hiddenimports:
    hiddenimports.append('ytmusicapi')

# aionowplaying is conditionally imported.
if os.name == 'nt':
    hiddenimports.append('aionowplaying')
    # pyinstaller can't detect 'aionowplaying.interface.windows' is imported.
    hiddenimports.append('aionowplaying.interface.windows')
if sys.platform == 'darwin':
    hiddenimports.append('aionowplaying.interface.macos')
    hiddenimports.append('feeluown.aionowplaying.macos')

# Collect feeluown's resource files, like icons, qss files, etc.
# Collect plugins' resource files.
datas += collect_data_files('feeluown', includes=assets_files)
for pkg in hiddenimports:
    datas += collect_data_files(pkg, includes=assets_files)


# Collect mpv dynamic-link library.
if os.name == 'nt':
    mpv_dylib = 'mpv-1.dll'
else:
    mpv_dylib = 'mpv'
mpv_dylib_path = ctypes.util.find_library(mpv_dylib)
binaries = [(mpv_dylib_path, '.'), ]



import datetime
from typing import List
from dataclasses import dataclass

from langchain.chat_models import init_chat_model
from langchain.agents import create_agent
from langchain.tools import tool, ToolRuntime
from langgraph.checkpoint.memory import InMemorySaver
from langchain_core.messages import BaseMessage
from langchain_core.callbacks import BaseCallbackHandler

from feeluown.app import App
from feeluown.library import (
    BriefSongModel,
    ModelState,
    get_standby_score,
    STANDBY_FULL_SCORE,
    STANDBY_DEFAULT_MIN_SCORE,
)
from feeluown.ai.prompt import generate_prompt_for_library
from feeluown.utils.dispatch import Signal


@dataclass
class AISongModel:
    """A song recommended by the AI.

    :param description: Recommendation reason or song description.
    """

    title: str
    artists_name: str
    description: str

    def to_brief_song(self) -> BriefSongModel:
        return BriefSongModel(
            state=ModelState.not_exists,
            source="ai",
            identifier="not-exists",
            title=self.title,
            artists_name=self.artists_name,
        )


def create_chat_model_with_config(config):
    return init_chat_model(
        config.OPENAI_MODEL,
        api_key=config.OPENAI_API_KEY,
        base_url=config.OPENAI_API_BASEURL,
        temperature=0,
        model_provider="openai",  # OpenAI compatible API
    )


@dataclass
class CopilotContext:
    copilot: "Copilot"
    app: "App"


@tool
def add_songs_to_playlist_candidates(
    songs: List[AISongModel], runtime: ToolRuntime
) -> bool:
    """Add songs to playlist candidates.

    :param songs: A list of AISongModel.
    """
    runtime.context.copilot.set_candidates(songs)


@tool
def play_song(song: AISongModel, runtime: ToolRuntime):
    """Play a song.

    :param song: A AISongModel.
    """
    runtime.context.app.playlist.play_model(song.to_brief_song())


tools = [add_songs_to_playlist_candidates, play_song]


def create_agent_with_config(config):
    model = create_chat_model_with_config(config)
    return create_agent(
        model=model,
        system_prompt="你是一个音乐播放器 AI 助手。",
        tools=tools,
        context_schema=CopilotContext,
        checkpointer=InMemorySaver(),
    )


def create_recommendation_agent_with_config(config):
    model = create_chat_model_with_config(config)
    return create_agent(
        model=model,
        system_prompt="你是一个音乐播放器 AI 助手。",
        tools=[add_songs_to_playlist_candidates],
        context_schema=CopilotContext,
    )


class AISongMatcher:
    def __init__(self, app: App):
        self._app = app

    async def match(self, ai_song: AISongModel) -> BriefSongModel:
        """Math a song by title and artists name.

        This API is in alpha stage.
        """
        origin = ai_song.to_brief_song()
        title, artists_name = ai_song.title, ai_song.artists_name
        candidates = []
        async for result in self._app.library.a_search(f"{title} {artists_name}"):
            if result is None:
                continue
            for standby in result.songs:
                score = get_standby_score(origin, standby)
                if score == STANDBY_FULL_SCORE:
                    return standby
                elif score >= STANDBY_DEFAULT_MIN_SCORE:
                    candidates.append((score, standby))
        sorted_candidates = sorted(candidates, key=lambda x: x[0], reverse=True)
        return sorted_candidates[0][1] if sorted_candidates else None


class AgentStreamCallback(BaseCallbackHandler):
    def __init__(self, copilot: "Copilot", *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__copilot = copilot

    def on_chain_start(self, serialized, inputs, **kwargs):
        self.__copilot.is_working = True

    def on_chain_end(self, outputs, **kwargs):
        self.__copilot.is_working = False

    def on_chain_error(self, error, **kwargs):
        self.__copilot.is_working = False


class Copilot:
    def __init__(self, app: App):
        self._app = app
        self._agent = create_agent_with_config(self._app.config)
        self._agent_context = CopilotContext(copilot=self, app=app)
        self._agent_stream_callback = AgentStreamCallback(self)
        self._candidates: List[AISongModel] = []
        self.candidates_changed = Signal()
        self._current_thread_id = 1
        # Agent is working or not
        # When the agent is streaming messages, it is working.
        self._is_working = False
        # emit(bool): working: true, not_working: false
        self.working_state_changed = Signal()

    @property
    def is_working(self) -> bool:
        return self._is_working

    @is_working.setter
    def is_working(self, working: bool):
        self._is_working = working
        self.working_state_changed.emit(working)

    def new_thread(self):
        self._current_thread_id += 1

    async def recommend_a_song(self) -> List[AISongModel]:
        """Create an adhoc agent to recommend a song.

        :return: A list of AISongModel.
        """
        agent = create_recommendation_agent_with_config(self._app.config)
        input = {
            "messages": [
                {"role": "system", "content": await generate_prompt(self._app)},
                {
                    "role": "system",
                    "content": (
                        "根据用户的音乐库收藏，分析用户的喜好，并且综合当前日期/时间等信息，"
                        "推荐1首合适的歌给用户，并将歌曲加入到播放列表候选中。"
                    ),
                },
            ]
        }
        await agent.ainvoke(
            input,
            context=self._agent_context,
        )
        return self._candidates

    async def astream_user_query(self, query: str):
        async for v in self._agent.astream(
            {"messages": [{"role": "user", "content": query}]},
            self.get_config(),
            stream_mode="messages",
            context=self._agent_context,
        ):
            yield v

    def set_candidates(self, ai_songs: List[AISongModel]):
        self._candidates = ai_songs
        self.candidates_changed.emit(ai_songs)

    def get_config(self):
        return {
            "configurable": {"thread_id": str(self._current_thread_id)},
            "callbacks": [self._agent_stream_callback],
        }

    def get_current_thread_history_messages(self) -> List[BaseMessage]:
        return self._agent.get_state(self.get_config()).values["messages"]


async def generate_prompt(app: App):
    """
    1. Today's date and current time.
    2. User's music library, including songs, albums, artists, playlists.
    """
    time_prompt = (
        f"当前时间是 {datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S')}。"
    )
    library_prompt = await generate_prompt_for_library(app.coll_mgr.get_coll_library())
    return f"{time_prompt}\n\n{library_prompt}"


if __name__ == "__main__":
    import os
    import asyncio

    from feeluown.debug import mock_app

    with mock_app() as app:
        app.config.OPENAI_API_KEY = os.environ.get("DEEPSEEK_API_KEY", "")
        app.config.OPENAI_API_BASEURL = "https://api.deepseek.com"
        app.config.OPENAI_MODEL = "deepseek-chat"

        copilot = Copilot(app)
        songs = asyncio.run(copilot.recommend_a_song())
        state = copilot._agent.get_state(copilot.get_config())



r"""
fuo request parser
~~~~~~~~~~~~~~~~~~~~

ps: Unlike common programming languages,
the fuo request syntax is more similar to shell command syntax.
In theory, we could refer to optparse to implement this parser,
but the original author (cosven) isn’t very interested in optparse and
is more enthusiastic about programming language parsers,
so the Parser was implemented in the following form.

Context-free grammar::

    expr: cmd (value)* (cmd_option)? (req_option)?
    req_option: REQ_DELIMETER options
    cmd_option: LBRACKET options RBRACKET
    options: ((option_expr)(COMMA option_expr)*)?
    option_expr: NAME (EQ value)
    values: (value)*
    value: NAME | STRING | INTEGER | FLOAT | FURI | UNQUOTE_STRING
    cmd: NAME

``search faint [artist="linkin park"]  #: json``

parse tree::

                                 expr
       /       /                  |                        \
      cmd    value            cmd_option                req_option
       |       |          /       |        \                 |      \
    'NAME'   STRING  LBRACKET option_expr RBRACKET   REQ_DELIMETER option_expr
       |       |        /    /    |     \      \              |        |
    'search' 'faint' '['  NAME   EQ   value    ']'           '#:'     NAME
                            |     |     |                              |
                        'artist' '='  STRING                         'json'
                                        |
                                   'linkin park'
"""

from typing import Any

from feeluown.server.excs import FuoSyntaxError
from feeluown.server.data_structure import Request
from .lexer import Lexer
from .lexer import (
    TOKEN_NAME, TOKEN_FURI, TOKEN_STRING, TOKEN_UNQUOTE_STRING,
    TOKEN_INTEGER, TOKEN_FLOAT, TOKEN_COMMA,
    TOKEN_LBRACKET, TOKEN_RBRACKET, TOKEN_EQ, TOKEN_REQ_DELIMETER,
    TOKEN_HEREDOC_OP, TOKEN_HEREDOC_WORD,
)


class _EOF(Exception):
    pass


class Parser:
    """fuo request syntax analyzer

    Implemented using recursive descent, top-down, LL(1).
    """

    def __init__(self, source):
        self.tokens = Lexer().tokenize(source)
        self._source = source

        self._end_column = len(self._source) - 1
        self._current_token: Any = None  # one lookahead token

    def parse(self):
        cmd = self._parse_cmd()
        req = Request(cmd)
        try:
            req.cmd_args = self._parse_values()
            req.cmd_options = self.parse_cmd_options()
            req.options = self._parse_req_options()
            req.has_heredoc, req.heredoc_word = self._parse_heredoc()
        except _EOF:
            # Encountering EOF in the above steps is normal.
            pass
        else:
            # Check if there are any remaining tokens
            try:
                token = self._next_token()
            except _EOF:
                pass
            else:
                self._error(token.column)
        return req

    def _error(self, column):
        raise FuoSyntaxError('invalid syntax',
                             column=column,
                             text=self._source)

    def _next_token(self):
        token = self._current_token
        if token is None:
            try:
                token = next(self.tokens)
            except StopIteration:
                raise _EOF from None
        self._current_token = None
        return token

    def _parse_cmd(self):
        """
        cmd: NAME

        :return: a value on success; on failure throws a FuoSyntaxError exception
        """
        try:
            token = self._next_token()
        except _EOF:
            raise FuoSyntaxError('invalid syntax: empty request') from None
        else:
            if token.type_ != TOKEN_NAME:
                self._error(token.column)
            return token.value

    def _parse_value(self):
        """
        value: NAME | STRING | INTEGER | FLOAT | FURI

        :return: a value on success; otherwise returns None.
        """
        valid_types = (TOKEN_NAME, TOKEN_STRING, TOKEN_INTEGER,
                       TOKEN_FLOAT, TOKEN_FURI, TOKEN_UNQUOTE_STRING)
        try:
            token = self._next_token()
        except _EOF:
            return None
        else:
            if token.type_ in valid_types:
                return token.value
            self._current_token = token
            return None

    def _parse_values(self):
        """
        values: (value)*
        """
        values = []
        while 1:
            value = self._parse_value()
            if value is None:
                break
            values.append(value)
        return values

    def _parse_option_expr(self):
        """
        option_expr: NAME (EQ value)

        :return: a (k, v) tuple on success; on failure returns (None, None)
        """
        token = self._next_token()
        if token.type_ == TOKEN_NAME:
            try:
                next_token = self._next_token()
            except _EOF:
                return token.value, True
            else:
                if next_token.type_ == TOKEN_EQ:
                    value = self._parse_value()
                    if value is None:
                        self._error(self._current_token.column)
                    elif value.lower() == 'true':
                        value = True
                    elif value.lower() == 'false':
                        value = False
                    return token.value, value
                self._current_token = next_token
                return token.value, True
        self._current_token = token
        return None, None

    def _parse_options(self):
        """
        options: ((option_expr)(COMMA option_expr)*)?
        """
        options = {}
        while 1:
            k, v = self._parse_option_expr()
            if k is None:
                break
            options[k] = v
            try:
                next_token = self._next_token()
            except _EOF:
                break
            else:
                if next_token.type_ == TOKEN_COMMA:
                    continue
            self._current_token = next_token
            break
        return options

    def parse_cmd_options(self):
        """
        cmd_options: LBRACKET ((option_expr)(COMMA option_expr)*)? RBRACKET
        """
        token = self._next_token()
        if token.type_ != TOKEN_LBRACKET:
            self._current_token = token
            return {}
        cmd_options = self._parse_options()
        next_token = self._next_token()  # Consume RBRACKET
        assert next_token.type_ == TOKEN_RBRACKET
        return cmd_options

    def _parse_req_options(self):
        next_token = self._next_token()  # Consume REQ_DELIMETER
        if next_token.type_ != TOKEN_REQ_DELIMETER:
            self._current_token = next_token
            return {}
        return self._parse_options()

    def _parse_heredoc(self):  # pylint: disable=inconsistent-return-statements
        token = self._next_token()
        if token.type_ != TOKEN_HEREDOC_OP:
            self._current_token = token
            return False, None
        try:
            next_token = self._next_token()
        except _EOF:
            self._error(self._end_column)
        else:
            assert next_token.type_ == TOKEN_HEREDOC_WORD
            return True, next_token.value


def parse(source):
    return Parser(source).parse()



"""
feeluown.app.config
~~~~~~~~~~~~~~~~~

App configuration items.
"""

from feeluown.config import Config
from feeluown.i18n import t


def create_config() -> Config:
    config = Config()
    config.deffield(
        "DEBUG",
        type_=bool,
        desc=t("debug-desc"),
    )
    config.deffield(
        "VERBOSE",
        default=0,
        type_=int,
        desc=t("verbose-desc"),
    )
    config.deffield(
        "RPC_PORT",
        default=23333,
        type_=int,
        desc=t("rpc-port-desc"),
    )
    config.deffield(
        "PUBSUB_PORT",
        default=23334,
        type_=int,
        desc=t("pubsub-port-desc"),
    )
    config.deffield(
        "WEB_PORT",
        default=23332,
        type_=int,
        desc=t("web-port-desc"),
    )
    config.deffield(
        "MCP_PORT",
        default=23335,
        type_=int,
    )
    config.deffield(
        "ENABLE_WEB_SERVER",
        default=False,
        type_=bool,
    )
    config.deffield(
        "ENABLE_MCP_SERVER",
        default=False,
        type_=bool,
    )
    config.deffield(
        "MODE",
        default=0x0000,
        desc=t("mode-desc"),
    )
    config.deffield(
        "THEME",
        default="auto",
        desc=t("theme-desc"),
    )
    config.deffield(
        "ENABLE_NEW_HOMEPAGE",
        default=True,
        type_=bool,
    )
    config.deffield(
        "NEW_HOMEPAGE_SETTINGS",
        type_=dict,
        default={
            "contents": [
                {"name": "RecListDailySongs", "provider": "netease"},
                # {'name': 'RecACollectionOfSongs', 'provider': 'qqmusic'},
                {"name": "RecListDailyPlaylists", "provider": "qqmusic"},
                {"name": "RecACollectionOfVideos", "provider": "bilibili"},
            ]
        },
        desc=t("new-homepage-settings-desc"),
    )
    config.deffield(
        "MPV_AUDIO_DEVICE",
        default="auto",
        desc=t("mpv-audio-device-desc"),
    )
    config.deffield(
        "COLLECTIONS_DIR",
        desc=t("collections-dir-desc"),
    )
    config.deffield(
        "FORCE_MAC_HOTKEY",
        desc=t("force-mac-hotkey-desc"),
        warn="Will be remove in version 3.0",
    )
    config.deffield(
        "LOG_TO_FILE",
        desc=t("log-to-file-desc"),
    )
    config.deffield(
        "AUDIO_SELECT_POLICY",
        default="hq<>",
    )
    config.deffield(
        "VIDEO_SELECT_POLICY",
        default="hd<>",
    )
    config.deffield(
        "ALLOW_LAN_CONNECT",
        type_=bool,
        default=False,
        desc=t("allow-lan-connect-desc"),
    )
    config.deffield(
        "PROVIDERS_STANDBY",
        type_=list,
        default=None,
        desc="",
    )

    # YTDL related fields are deprecated since v4.1.9. Disable them by default.
    config.deffield(
        "ENABLE_YTDL_AS_MEDIA_PROVIDER",
        type_=bool,
        default=False,
        desc=t("enable-ytdl-as-media-provider-desc"),
    )
    # For example::
    #    [
    #        {
    #            'name': 'match_source',
    #            'pattern': 'ytmusic',
    #            'http_proxy': 'http://127.0.0.1:7890',
    #            'ytdl_options': {
    #                'socket_timeout': 2,
    #            },
    #        },
    #    ]
    config.deffield(
        "YTDL_RULES",
        type_=list,
        default=None,
        desc=t("ytdl-rules-desc"),
    )

    # TODO(cosven): maybe
    # 1. when it is set to 2, find standby from other providers first.
    # 2. when it is set to 3, play it's MV model instead of using MV's media.
    config.deffield(
        "ENABLE_MV_AS_STANDBY",
        type_=int,
        default=1,
        desc=t("enable-mv-as-standby-desc"),
    )
    config.deffield(
        "ENABLE_TRAY",
        type_=bool,
        default=True,
        desc=t("enable-tray-desc"),
    )
    config.deffield(
        "NOTIFY_ON_TRACK_CHANGED",
        type_=bool,
        default=False,
        desc=t("notify-on-track-changed-desc"),
    )
    config.deffield(
        "NOTIFY_DURATION",
        type_=int,
        default=3000,
        desc=t("notify-duration-desc"),
    )
    config.deffield(
        "PLAYBACK_CROSSFADE",
        type_=bool,
        default=False,
        desc=t("playback-crossfade-desc"),
    )
    config.deffield(
        "PLAYBACK_CROSSFADE_DURATION",
        type_=int,
        default=500,
        desc=t("playback-crossfade-desc"),
    )
    config.deffield(
        "OPENAI_API_BASEURL",
        type_=str,
        default="",
        desc=t("openai-api-baseurl-desc"),
    )
    config.deffield(
        "OPENAI_API_KEY",
        type_=str,
        default="",
        desc=t("openai-api-key-desc"),
    )
    config.deffield(
        "OPENAI_MODEL",
        type_=str,
        default="",
        desc=t("openai-model-desc"),
    )
    # Not sure if AI_STANDBY_MATCHER may be activated unexpectedly.
    # And it may burn some money, let user enable it manually.
    config.deffield(
        "ENABLE_AI_STANDBY_MATCHER",
        type_=bool,
        default=False,
        desc="",
    )
    config.deffield(
        "AI_RADIO_PROMPT",
        type_=str,
        default="你是一个音乐播放器智能助手。",
        desc=t("ai-radio-prompt-desc"),
    )
    config.deffield(
        "ENABLE_REPLACE_PLAYLIST_ON_DBLCLICK",
        type_=bool,
        default=True,
        desc=t("enable-replace-playlist-on-dblclick-desc"),
    )
    return config



#! /usr/bin/env python3
# -*- coding: utf-8 -*-

"""
You should install two package from pypi:

1. fuo-netease
2. fuo-xiami
"""

import logging
from feeluown.library import Library

from fuo_xiami import provider as xp
from fuo_netease import provider as np

logging.basicConfig()
logger = logging.getLogger('feeluown')
logger.setLevel(logging.DEBUG)

lib = Library()
lib.register(xp)
lib.register(np)


def test_list_song_standby():
    """
    Test the library.list_song_standby function
    """
    result = xp.search('小小恋歌 新垣结衣', limit=2)
    song = result.songs[0]
    assert song.url == ''

    standby_songs = lib.list_song_standby(song)
    for index, ss in enumerate(standby_songs):  # pylint: disable=all
        print(index, ss.source, ss.title, ss.artists_name, ss.url)


def main():
    test_list_song_standby()


if __name__ == '__main__':
    main()



from collections import defaultdict
from typing import Any
from unittest.mock import Mock

from feeluown.app import App
from feeluown.player import Metadata
from feeluown.library import (
    BaseModel,
    SongModel,
    ArtistModel,
    AlbumModel,
    PlaylistModel,
    UserModel,
    BriefSongModel,
    BriefArtistModel,
    BriefAlbumModel,
    BriefPlaylistModel,
    BriefUserModel,
)

model_cls_list = [
    BaseModel,
    SongModel,
    ArtistModel,
    AlbumModel,
    PlaylistModel,
    UserModel,
    BriefSongModel,
    BriefArtistModel,
    BriefAlbumModel,
    BriefPlaylistModel,
    BriefUserModel,
]

_sys_typenames = {  # for unittest.
    'unittest.mock.Mock': Mock
}
_typenames = {
    'player.Metadata': Metadata,
    'app.App': App,  # TODO: remove this
}
for model_cls in model_cls_list:
    _typenames[f'library.{model_cls.__name__}'] = model_cls
typenames = {f'feeluown.{k}': v for k, v in _typenames.items()}
typenames.update(_sys_typenames)

r_typenames = defaultdict(list)
for k, v in typenames.items():
    r_typenames[v].append(k)
r_typenames = dict(r_typenames)


def get_type_by_name(name: str):
    return typenames.get(name, None)


def get_names_by_type(type_: Any):
    # try except so that performance is not affected.
    try:
        return r_typenames[type_]
    except KeyError:
        if type_.__module__ == 'unittest.mock':
            return ['unittest.mock.Mock']
        return []


def attach_typename(method):

    def wrapper(this, obj, **kwargs):
        result = method(this, obj, **kwargs)
        if isinstance(result, dict):
            typenames = get_names_by_type(type(obj))
            if typenames:
                result['__type__'] = typenames[0]
            else:
                result['__type__'] = 'to_be_added__pr_is_welcome'
        return result

    return wrapper



import asyncio
import logging
import socket


logger = logging.getLogger(__name__)


class TcpServer(object):
    """A simple asyncio TCP server

    This implementation is based on low-level event loop APIs such as
    `loop.sock_accept`, `loop.sock_sendall`, and `loop.sock_recv`.

    Asyncio also provides `loop.create_server` for creating a server, together with
    `StreamReader` and `StreamWriter` for I/O, which is overall a more high-level
    abstraction. `StreamReader` wraps convenient methods such as `readline` and
    `readuntil`, making it easier to use.

    `StreamWriter` maintains its own internal buffer, so calling `write()` does not
    immediately send data to the operating system. Prior to Python 3.7, there was
    no particularly good way to reliably ensure this behavior. However, it does
    provide a `drain()` method to flush part of the buffered data to the OS.

    Overall, I personally prefer this `TcpServer` implementation. Its interface is
    very explicit and feels much closer to using blocking sockets. The asyncio
    stream-based API, on the other hand, is less intuitive to me. That said, the
    current implementation in FeelUOwn does use asyncio streams—probably due to a
    bit of personal perfectionism.
    """

    def __init__(self, host, port, handle_func):
        self.host = host
        self.port = port
        self.handle_func = handle_func

        self._sock = None

    async def run(self, *args, **kwargs):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind((self.host, self.port))
        sock.listen()
        # make restart easier
        sock.setblocking(0)
        self._sock = sock

        event_loop = asyncio.get_event_loop()
        while True:
            conn, addr = await event_loop.sock_accept(sock)
            event_loop.create_task(
                self.handle_func(conn, addr, *args, **kwargs))



from datetime import datetime

from PyQt6.QtCore import QAbstractListModel, QModelIndex, Qt, QSize, QPoint, QRect
from PyQt6.QtGui import QPalette, QPen, QFontMetrics
from PyQt6.QtWidgets import QStyledItemDelegate, QListView, QSizePolicy, QFrame, QWidget

from feeluown.gui.consts import ScrollBarWidth
from feeluown.gui.helpers import ItemViewNoScrollMixin, Paddings, Margins
from feeluown.library import CommentModel
from feeluown.utils.reader import Reader

from feeluown.i18n import human_readable_number


class CommentListModel(QAbstractListModel):
    def __init__(self, reader: Reader[CommentModel], parent=None):
        super().__init__(parent)

        self._reader = reader

    def rowCount(self, _=QModelIndex()):
        return self._reader.count

    def columnCount(self, _=QModelIndex()):
        return 1

    def flags(self, index):
        if not index.isValid():
            return 0
        flags = Qt.ItemFlag.ItemIsSelectable | Qt.ItemFlag.ItemIsEnabled
        return flags

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        if role == Qt.ItemDataRole.UserRole:
            return self._reader.read(index.row())
        return None


class CommentListDelegate(QStyledItemDelegate):
    def __init__(self, parent: QWidget, quoted_bg_color_role=QPalette.ColorRole.Window):
        super().__init__(parent=parent)

        # macOS scrollbar may overlap with the content.
        # So set the margin_h_right to `ScrollBarWidth * 2`.
        self._margin_h_left, self._margin_h_right = 0, ScrollBarWidth * 2
        self._margin_v = 10
        self._name_content_margin = 5
        self._quoted_bg_color_role = quoted_bg_color_role
        self._name_height = QFontMetrics(parent.font()).height()
        self._parent_comment_paddings = Paddings(8, 3, 8, 3)
        self._parent_comment_margins = Margins(20, 5, 10, 5)

    @property
    def _margin_h_total(self):
        return self._margin_h_left + self._margin_h_right

    def paint(self, painter, option, index):
        # pylint: disable=too-many-locals,too-many-statements
        # Explicitly set the font to the same font, so that the content
        # height is always correct. Otherwise, the height maybe incorrect.
        fm = QFontMetrics(option.font)

        painter.save()
        comment = index.data(Qt.ItemDataRole.UserRole)

        # size for render comment
        body_width = option.rect.width() - self._margin_h_total
        body_height = option.rect.height() - self._margin_v * 2

        painter.translate(
            QPoint(
                option.rect.x() + self._margin_h_left, option.rect.y() + self._margin_v
            )
        )

        # draw comment author name
        painter.save()
        font = option.font
        font.setBold(True)
        painter.setFont(font)
        name_rect = QRect(0, 0, body_width, self._name_height)
        painter.drawText(name_rect, Qt.AlignmentFlag.AlignLeft, comment.user.name)
        painter.restore()

        # draw comment other metadata
        painter.save()
        metadata_rect = QRect(0, 0, body_width, self._name_height)
        text_list = []
        if comment.time:
            dt = datetime.fromtimestamp(comment.time)
            text_list.append(dt.strftime("%Y-%m-%d %H:%M"))
        if comment.liked_count != -1:
            liked_count_text = human_readable_number(comment.liked_count)
            text_list.append(f"♥ {liked_count_text}")
        text = "  |  ".join(text_list)
        text_color = option.palette.color(QPalette.ColorRole.Text)
        text_color.setAlpha(100)
        pen = QPen()
        pen.setColor(text_color)
        painter.setPen(pen)
        if text:
            painter.drawText(metadata_rect, Qt.AlignmentFlag.AlignRight, text)
        painter.restore()

        # draw comment content
        painter.save()
        name_height = self._name_height + self._name_content_margin
        painter.translate(QPoint(0, name_height))
        content_height = self._get_text_height(fm, body_width, comment.content)
        content_rect = QRect(0, 0, body_width, content_height)
        painter.drawText(content_rect, Qt.TextFlag.TextWordWrap, comment.content)
        parent_comment = comment.parent
        if parent_comment is not None:
            p_margins = self._parent_comment_margins
            p_paddings = self._parent_comment_paddings
            text = f"{parent_comment.user_name}：{parent_comment.content}"
            p_width = body_width - p_margins.width
            p_height = self._get_text_height(fm, p_width, text)
            p_body_rect = QRect(
                p_margins.left,
                p_margins.top + content_height,
                p_width,
                p_height + p_paddings.height,
            )
            p_content_rect = QRect(
                p_body_rect.x() + p_paddings.left,
                p_body_rect.y() + p_paddings.top,
                p_body_rect.width() - p_paddings.width,
                p_body_rect.height() - p_paddings.height,
            )
            bg_color = option.palette.color(self._quoted_bg_color_role)
            painter.fillRect(p_body_rect, bg_color)
            painter.drawText(p_content_rect, Qt.TextFlag.TextWordWrap, text)

        painter.restore()

        # draw a dotted line under each row
        painter.save()
        painter.translate(QPoint(0, body_height + self._margin_v))
        text_color = option.palette.color(QPalette.ColorRole.Text)
        if text_color.lightness() > 150:
            non_text_color = text_color.darker(140)
        else:
            non_text_color = text_color.lighter(150)
        non_text_color.setAlpha(30)
        painter.setPen(QPen(non_text_color, 1, Qt.PenStyle.DotLine))
        painter.drawLine(QPoint(0, 0), QPoint(body_width, 0))
        painter.restore()

        painter.restore()

    def sizeHint(self, option, index):
        super_size_hint = super().sizeHint(option, index)
        parent_width = self.parent().width()  # type: ignore
        fm = QFontMetrics(option.font)
        comment = index.data(Qt.ItemDataRole.UserRole)
        content_width = parent_width - self._margin_h_total
        content_height = self._get_text_height(fm, content_width, comment.content)
        height = (
            content_height
            + self._name_height
            + self._name_content_margin
            + self._margin_v * 2
        )
        parent_comment = comment.parent
        if parent_comment is not None:
            p_margins = self._parent_comment_margins
            p_paddings = self._parent_comment_paddings
            text = f"{parent_comment.user_name}：{parent_comment.content}"
            p_height = self._get_text_height(fm, content_width - p_margins.width, text)
            height += p_height + p_margins.height + p_paddings.height
        return QSize(super_size_hint.width(), height)

    def _get_text_height(self, fm, width, text):
        return fm.boundingRect(
            QRect(0, 0, width, 0), Qt.TextFlag.TextWordWrap, text
        ).height()


class CommentListView(ItemViewNoScrollMixin, QListView):
    def __init__(self, parent=None, **kwargs):
        delegate_options = kwargs.pop("delegate_options", {})
        super().__init__(parent=parent, **kwargs)

        self._delegate = CommentListDelegate(self, **delegate_options)
        self.setItemDelegate(self._delegate)
        self.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Expanding)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        self.setMouseTracking(True)
        self.setFrameShape(QFrame.Shape.NoFrame)
        self.setResizeMode(QListView.ResizeMode.Adjust)

    def min_height(self):
        """
        override ItemViewNoScrollMixin.min_height
        """
        model = self.model()
        if model is None:
            return super().min_height()
        row_count = model.rowCount()
        height = 0
        while row_count >= 0:
            row_count -= 1
            height += self.sizeHintForRow(row_count)
        return height



import logging

from PyQt6.QtCore import QSize, Qt, QRectF, pyqtSignal
from PyQt6.QtGui import QPainter, QPainterPath, QTextOption
from PyQt6.QtWidgets import (
    QFrame,
    QHBoxLayout,
    QLabel,
    QPlainTextEdit,
    QScrollArea,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)

from feeluown.i18n import t
from feeluown.gui.helpers import palette_set_bg_color
from feeluown.gui.widgets.textbtn import TextButton

logger = logging.getLogger(__name__)


class ChatInputEditor(QPlainTextEdit):
    """Custom editor for chat input with Enter key handling"""

    enter_pressed = pyqtSignal()

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWordWrapMode(QTextOption.WrapMode.WrapAtWordBoundaryOrAnywhere)
        self.setMinimumHeight(30)
        self.setMaximumHeight(300)  # TODO: set maximum height based on parent size
        self.textChanged.connect(self.adjust_height)
        self.adjust_height()
        # The size policy matters
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Preferred)

    def sizeHint(self) -> QSize:  # noqa: D102
        font_metrics = self.fontMetrics()
        line_height = font_metrics.lineSpacing()
        doc = self.document()
        line_count = doc.lineCount()
        doc_height = line_count * line_height

        # Add some padding, 10
        new_height = min(
            max(int(doc_height) + 10, self.minimumHeight()), self.maximumHeight()
        )
        return QSize(super().sizeHint().width(), new_height)

    def adjust_height(self):
        self.updateGeometry()

    def keyPressEvent(self, event):
        if event.key() == Qt.Key.Key_Return and not event.modifiers():
            self.enter_pressed.emit()
            event.accept()
        else:
            super().keyPressEvent(event)


class RoundedLabel(QLabel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._radius = 8
        self._padding = 8
        self.setContentsMargins(
            self._padding, self._padding, self._padding, self._padding
        )
        self.setWordWrap(True)

    def paintEvent(self, event):
        painter = QPainter(self)
        path = QPainterPath()
        path.addRoundedRect(QRectF(self.rect()), self._radius, self._radius)
        # Fill background
        painter.fillPath(path, self.palette().color(self.backgroundRole()))
        super().paintEvent(event)


class ChatHistoryWidget(QWidget):
    """Widget for displaying chat history"""

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self._history_area = QScrollArea(self)
        self.history_widget = QWidget()
        self._history_area.setWidget(self.history_widget)
        self._history_layout = QVBoxLayout(self.history_widget)

        self._history_area.setFrameShape(QFrame.Shape.NoFrame)
        self._history_area.setAutoFillBackground(True)
        self._history_layout.setContentsMargins(0, 0, 20, 0)
        self._history_layout.setSpacing(5)
        self._history_area.setWidgetResizable(True)
        # Adjust spacing between messages
        self._history_layout.setSpacing(10)
        self._history_layout.addStretch(0)

        layout = QVBoxLayout(self)
        layout.addWidget(self._history_area)
        layout.setContentsMargins(0, 0, 0, 0)

    def add_message(self, role, content):
        """Add a message to the history"""
        self.create_message_label(role, content)

    def create_message_label(self, role, content):
        """Create message label"""
        label = RoundedLabel()
        label.setText(content)
        label.setWordWrap(True)
        label.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse)
        label.setFrameStyle(QFrame.Shape.NoFrame)

        label.setSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred)
        label.setAlignment(Qt.AlignmentFlag.AlignLeft)

        pal = label.palette()
        if role in ("user", "system"):
            origin_window = pal.color(pal.ColorRole.Window)
            palette_set_bg_color(pal, pal.color(pal.ColorRole.Highlight))
            pal.setColor(pal.ColorRole.Text, pal.color(pal.ColorRole.HighlightedText))
            pal.setColor(pal.ColorRole.Highlight, origin_window)
            label.setPalette(pal)
        elif role in ("tools"):
            font = label.font()
            font.setPixelSize(11)
            font.setFamilies(["Monaco", "Menlo", "monospace"])
            label.setFont(font)

            pal.setColor(pal.ColorRole.Text, pal.color(pal.ColorRole.ToolTipText))
            palette_set_bg_color(pal, pal.color(pal.ColorRole.ToolTipBase))
            label.setPalette(pal)

        self._history_layout.addWidget(label)
        self.scroll_to_bottom()
        return label

    def scroll_to_bottom(self) -> None:
        """Scroll chat history to bottom."""
        self.history_widget.updateGeometry()
        self._history_area.verticalScrollBar().setValue(
            self._history_area.verticalScrollBar().maximum()
        )

    def clear(self):
        """Clear chat history"""
        while self._history_layout.count():
            item = self._history_layout.takeAt(0)
            widget = item.widget()
            if widget:
                widget.deleteLater()

    def widget(self):
        """Return the widget to be added to layouts"""
        return self


class ChatInputWidget(QWidget):
    """Widget for chat input and buttons"""

    send_clicked = pyqtSignal(str)  # emits query text

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self._editor = ChatInputEditor(self)
        self._editor.setPlaceholderText(t("ai-chat-input-placeholder"))
        self._editor.setFrameShape(QFrame.Shape.NoFrame)
        self._msg_label = QLabel(self)
        self._msg_label.setWordWrap(True)
        self._send_btn = TextButton(t("ai-chat-send-button"))

        self._editor.enter_pressed.connect(self._on_enter_pressed)
        self._send_btn.clicked.connect(self._on_send_clicked)
        self._editor.setMinimumHeight(80)
        self.setup_ui()

    def setup_ui(self):
        self._msg_label.setWordWrap(True)
        self._msg_label.setTextFormat(Qt.TextFormat.RichText)

        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(5)
        layout.addWidget(self._msg_label)

        input_layout = QHBoxLayout()
        input_layout.addWidget(self._editor)
        input_layout.addWidget(self._send_btn, 0, Qt.AlignmentFlag.AlignBottom)
        layout.addLayout(input_layout)

    def _on_enter_pressed(self):
        self._on_send_clicked()

    def _on_send_clicked(self):
        query = self._editor.toPlainText().strip()
        if query:
            self.send_clicked.emit(query)
            self._editor.clear()

    def set_msg(self, text, level="hint"):
        if level == "hint":
            color = "green"
        elif level == "warn":
            color = "yellow"
        else:  # err
            color = "magenta"
        self._msg_label.setText(f'<span style="color: {color}">{text}</span>')

    def clear_input(self):
        self._editor.clear()

    def get_input(self):
        return self._editor.toPlainText()

    def enable_send(self, enabled):
        self._send_btn.setEnabled(enabled)



"""
A component is a certain widget depends on `feeluown.app.App` object.
In other words, a component is tightly coupled with application logic.
"""

from .avatar import Avatar  # noqa
from .btns import *  # noqa
from .menu import SongMenuInitializer  # noqa
from .line_song import LineSongLabel  # noqa
from .playlist_btn import PlaylistButton  # noqa
from .volume_slider import *  # noqa
from .song_tag import SongSourceTag  # noqa
from .collections import CollectionListView  # noqa
from .player_progress import PlayerProgressSliderAndLabel  # noqa
from .recommendation_panel import Panel, RecPlaylistsPanel  # noqa



import math

from PyQt6.QtCore import Qt, QSize, QRectF, QRect, QPoint
from PyQt6.QtGui import QPainter, QTextOption, QPalette, QBrush, QColor, QGuiApplication  # noqa
from PyQt6.QtWidgets import QStyledItemDelegate, QListView
from PyQt6.QtSvg import QSvgRenderer

from feeluown.library import ModelType, ProviderFlags as PF
from feeluown.gui.helpers import ItemViewNoScrollMixin, SOLARIZED_COLORS
from .textlist import TextlistModel

from feeluown.i18n import t


class ProvidersModel(TextlistModel):
    def __init__(self, library, parent=None):
        super().__init__(parent)
        self._library = library
        self._association = {}

    def assoc(self, provider_id, pm):
        self._association[provider_id] = pm
        self.add(provider_id)

    def remove(self, provider_id):
        if not self._association.pop(provider_id, None):
            self.remove(provider_id)

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        row = index.row()
        provider_id = self._items[row]
        provider_ui_item = self._association[provider_id]
        if role == Qt.ItemDataRole.DisplayRole:
            return provider_ui_item.symbol + " " + provider_ui_item.text
        if role == Qt.ItemDataRole.ToolTipRole:
            if self._library.check_flags(provider_id, ModelType.none, PF.current_user):
                provider = provider_ui_item.provider
                if provider.has_current_user():
                    return t("logged")
                return f"{provider_ui_item.desc}"
            return f"[{provider_ui_item.text}] {provider_ui_item.desc}"
        if role == Qt.ItemDataRole.UserRole:
            return provider_ui_item
        return super().data(index, role)


class ProvidersDelegate(QStyledItemDelegate):
    def __init__(self, parent=None, library=None):
        super().__init__(parent)
        self._library = library

        self._radius = 24
        self._padding = 7

        self.__body_radius = self._radius - self._padding
        self.__text_rect_x = self.__body_radius - self.__body_radius / math.sqrt(2)
        self.__text_rect_width = 2 * (self.__body_radius / math.sqrt(2))

    def paint(self, painter, option, index):
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)
        provider_ui_item = index.data(Qt.ItemDataRole.UserRole)

        painter.save()
        painter.translate(
            self._padding + option.rect.x(), self._padding + option.rect.y()
        )

        w = h = (self._radius - self._padding) * 2
        body_rect = QRect(0, 0, w, h)

        if provider_ui_item.colorful_svg:
            svg_renderer = QSvgRenderer(provider_ui_item.colorful_svg)
            svg_renderer.render(painter, QRectF(body_rect))
        else:
            # draw rounded rect
            painter.save()
            text_color = option.palette.color(QPalette.ColorRole.Text)
            if text_color.lightness() > 150:
                non_text_color = text_color.darker(140)
            else:
                non_text_color = text_color.lighter(150)
            non_text_color.setAlpha(100)
            pen = painter.pen()
            pen.setColor(non_text_color)
            painter.setPen(pen)
            painter.drawRoundedRect(body_rect, w // 2, h // 2)
            painter.restore()

            painter.save()
            font = painter.font()
            font.setPixelSize(10)
            painter.setFont(font)
            text_option = QTextOption()
            text_option.setWrapMode(QTextOption.WrapMode.WrapAtWordBoundaryOrAnywhere)
            text_option.setAlignment(Qt.AlignmentFlag.AlignCenter)

            text_rect = QRectF(
                self.__text_rect_x,
                self.__text_rect_x,
                self.__text_rect_width,
                self.__text_rect_width,
            )
            painter.drawText(QRectF(text_rect), provider_ui_item.text, text_option)
            painter.restore()

        # TODO: use library.check_flags instead of provider.check_flags
        provider = provider_ui_item.provider
        if self._library.check_flags(
            provider.identifier, ModelType.none, PF.current_user
        ):
            if provider.has_current_user():
                painter.save()
                bottom_right = body_rect.bottomRight()
                status_radius = self._radius // 5
                x = bottom_right.x() - status_radius * 2
                y = bottom_right.y() - status_radius * 2
                status_rect = QRect(QPoint(x, y), bottom_right)
                pen = painter.pen()
                pen.setWidth(2)
                pen.setColor(QGuiApplication.palette().color(QPalette.ColorRole.Window))
                painter.setPen(pen)
                painter.setBrush(QColor(SOLARIZED_COLORS["blue"]))
                painter.drawRoundedRect(status_rect, status_radius, status_radius)
                painter.restore()

        painter.restore()

    def sizeHint(self, option, index):
        return QSize(self._radius * 2, self._radius * 2)


class ProvidersView(ItemViewNoScrollMixin, QListView):
    def __init__(self, parent, library):
        super().__init__(parent=parent)
        self._library = library

        self._least_row_count = 1
        self._row_height = 48

        self.delegate = ProvidersDelegate(self, library=library)
        self.setItemDelegate(self.delegate)
        self.setViewMode(QListView.ViewMode.IconMode)
        self.setResizeMode(QListView.ResizeMode.Adjust)
        self.setWrapping(True)

        self.clicked.connect(self._on_clicked)

    def _on_clicked(self, index):
        provider_ui_item = index.data(role=Qt.ItemDataRole.UserRole)
        provider_ui_item.clicked.emit()

    def resizeEvent(self, e):
        super().resizeEvent(e)
        self.adjust_height()



import re
from collections import namedtuple
from urllib.parse import parse_qsl, urlsplit


class NotFound(Exception):
    """
    No matched rule for path
    """


def match(url, rules):
    """Find the rule corresponding to the path and parse its parameters.

    >>> match('/local/songs', rules=['/<p>/songs'])
    ('/<p>/songs', {'p': 'local'}, {})
    >>> match('/search?q=hello+world', rules=['/search'])
    ('/search', {}, {'q': 'hello world'})

    :return: (rule, params) or None
    """
    split_result = urlsplit(url)
    path = split_result.path
    qs = split_result.query
    for rule in rules:
        url_regex = regex_from_rule(rule)
        match = url_regex.match(path)
        if match:
            # The result of parse_qsl may be [('a', 'b'), ('a', 'c')],
            # corresponding to the query string a=b&a=c.
            # We currently do not allow this case, so here we
            # temporarily convert the parsed results directly into a dict
            query = dict(parse_qsl(qs))
            params = match.groupdict()
            return rule, params, query
    raise NotFound(f"No matched rule for {path}")


def _validate_rule(rule):
    """Simple validation of rule

    TODO: code implementation needs improvement
    """
    if rule:
        if rule == '/':
            return
        parts = rule.split('/')
        if parts.count('') == 1:
            return
    raise ValueError('Invalid rule: {}'.format(rule))


def regex_from_rule(rule):
    r"""为一个 rule 生成对应的正则表达式

    >>> regex_from_rule('/<provider>/songs')
    re.compile('^/(?P<provider>[^\\/]+)/songs$')
    """
    kwargs_regex = re.compile(r'(<.*?>)')
    pattern = re.sub(
        kwargs_regex,
        lambda m: r'(?P<{}>[^\/]+)'.format(m.group(0)[1:-1]),
        rule
    )
    regex = re.compile(r'^{}$'.format(pattern))
    return regex


Request = namedtuple('Request', ['uri', 'rule', 'params', 'query', 'ctx'])


class Router(object):
    def __init__(self):
        self.rules = []
        self.handlers = {}

    def register(self, rule, handler):
        self.rules.append(rule)
        # TODO: Perhaps we could use weakref?
        self.handlers[rule] = handler

    def route(self, rule):
        """show handler router decorator

        example::

            @route('/<provider_name>/songs')
            def show_songs(provider_name):
                pass
        """
        _validate_rule(rule)

        def decorator(func):
            self.register(rule, func)

            def wrapper(*args, **kwargs):
                func(*args, **kwargs)
            return wrapper
        return decorator

    def dispatch(self, uri, ctx):
        rule, params, query = match(uri, self.rules)
        handler = self.handlers[rule]
        req = Request(uri=uri, rule=rule, params=params, query=query, ctx=ctx)
        return handler(req, **params)



# Old code imports these exceptions from this module
from feeluown.excs import (  # noqa
    LibraryException,
    ResourceNotFound,
    # FIXME: ProviderAlreadyExists should be renamed to ProviderAlreadyRegistered
    ProviderAlreadyRegistered as ProviderAlreadyExists,
    ModelNotFound,
    NoUserLoggedIn,
    MediaNotFound,
)  # noqa



from feeluown.library import SearchType
from feeluown.utils.router import Request
from feeluown.app.gui_app import GuiApp
from feeluown.gui.components.search import SearchResultView


def get_source_in(req: Request):
    source_in = req.query.get("source_in", None)
    if source_in is not None:
        source_in = source_in.split(",")
    else:
        source_in = None
    return source_in


async def render(req: Request, **kwargs):
    """/search handler

    :type app: feeluown.app.App
    """
    # pylint: disable=too-many-locals,too-many-branches
    q = req.query.get("q", "")
    if not q:
        return

    app: "GuiApp" = req.ctx["app"]
    source_in = get_source_in(req)
    search_type = SearchType(req.query.get("type", SearchType.so.value))

    view = SearchResultView(app)
    app.ui.right_panel.set_body(view)
    await view.search_and_render(q, search_type, source_in)



# pylint: disable=import-error
import logging
import os

import aionowplaying as aionp

from feeluown.app.server_app import ServerApp
from feeluown.player import State, PlaybackMode

logger = logging.getLogger(__name__)
PlayProp = aionp.PlaybackPropertyName
StatePlaybackStatusMapping = {
    State.stopped: aionp.PlaybackStatus.Stopped,
    State.paused: aionp.PlaybackStatus.Paused,
    State.playing: aionp.PlaybackStatus.Playing,
}
PlaybackStatusStateMapping = {
    v: k
    for k, v in StatePlaybackStatusMapping.items()
}


def sec_to_us(t_sec):
    return int(t_sec * 1000 * 1000)


def us_to_sec(t_microsec):
    return t_microsec / 1000 / 1000


class NowPlayingService(aionp.NowPlayingInterface):
    def __init__(self, app: 'ServerApp'):
        super().__init__('FeelUOwn')
        self._app = app

        if os.name == "nt":
            self._app.player.position_changed.connect(self.update_position)

        self._app.player.seeked.connect(self.update_position)
        self._app.player.media_loaded.connect(self.on_player_media_loaded,
                                              aioqueue=True)
        self._app.player.duration_changed.connect(self.update_duration,
                                                  aioqueue=True)
        self._app.player.state_changed.connect(self.update_playback_status)
        self._app.player.metadata_changed.connect(self.update_song_metadata,
                                                  aioqueue=True)
        self._app.player.media_changed.connect(self.on_player_media_changed,
                                               aioqueue=True)
        self._app.playlist.playback_mode_changed.connect(
            self.update_playback_mode)
        self._app.started.connect(lambda: self.update_playback_mode(
            self._app.playlist.playback_mode))

        self.set_playback_property(PlayProp.CanPlay, True)
        self.set_playback_property(PlayProp.CanPause, True)
        self.set_playback_property(PlayProp.CanGoNext, True)
        self.set_playback_property(PlayProp.CanGoPrevious, True)
        self.set_playback_property(PlayProp.CanSeek, True)
        self.set_playback_property(PlayProp.CanControl, True)
        self.set_playback_property(PlayProp.Rate, 1.0)

    def update_playback_mode(self, mode: PlaybackMode):
        mode_mapping = {
            PlaybackMode.loop: (aionp.LoopStatus.Playlist, False),
            PlaybackMode.one_loop: (aionp.LoopStatus.Track, False),
            PlaybackMode.random: (aionp.LoopStatus.Playlist, True),
            PlaybackMode.sequential: (aionp.LoopStatus.None_, False),
        }
        mode_value = mode_mapping.get(mode)
        self.set_playback_property(PlayProp.LoopStatus, mode_value[0])
        self.set_playback_property(PlayProp.Shuffle, mode_value[1])

    def update_song_metadata(self, meta: dict):
        metadata = aionp.PlaybackProperties.MetadataBean()
        metadata.artist = meta.get('artists', ['Unknown'])
        metadata.album = meta.get('album', '')
        metadata.title = meta.get('title', '')
        metadata.cover = meta.get('artwork', '')
        metadata.url = meta.get('uri', '')
        if os.name == 'nt':
            metadata.id_ = meta.get('uri', '')
        self.set_playback_property(PlayProp.Metadata, metadata)

    def update_duration(self, duration):
        self.set_playback_property(PlayProp.Duration, int(sec_to_us(duration or 0)))

    def update_position(self, position):
        self.set_playback_property(PlayProp.Position, int(sec_to_us(position or 0)))

    def update_playback_status(self, state):
        self.set_playback_property(PlayProp.PlaybackStatus,
                                   StatePlaybackStatusMapping[state])

    async def on_play(self):
        self._app.player.resume()

    async def on_pause(self):
        self._app.player.pause()

    async def on_next(self):
        self._app.playlist.next()

    async def on_previous(self):
        self._app.playlist.previous()

    def on_stop(self):
        self._app.player.stop()

    def on_loop_status(self, status: aionp.LoopStatus):
        if status == aionp.LoopStatus.Playlist:
            self._app.playlist.playback_mode = PlaybackMode.loop
        elif status == aionp.LoopStatus.Track:
            self._app.playlist.playback_mode = PlaybackMode.one_loop
        else:
            self._app.playlist.playback_mode = PlaybackMode.sequential

    def on_shuffle(self, shuffle: bool):
        if shuffle:
            self._app.playlist.playback_mode = PlaybackMode.random
        else:
            self._app.playlist.playback_mode = PlaybackMode.sequential

    def on_set_position(self, track_id: str, position: int):
        self._app.player.position = us_to_sec(position)

    def on_seek(self, offset_us: int):
        self._app.player.position = us_to_sec(offset_us)

    def get_playback_property(self, name: PlayProp):
        if name == PlayProp.Duration:
            return sec_to_us(self._app.player.duration)
        elif name == PlayProp.Position:
            return sec_to_us(self._app.player.position)
        elif name == PlayProp.PlaybackStatus:
            return StatePlaybackStatusMapping[self._app.player.state]
        elif name == PlayProp.Rate:
            return 1.0
        else:
            raise ValueError('unknown key')

    def on_player_media_changed(self, _):
        # Update position when media is changed, so that some nowplaying servers
        # can update its UI immediately.
        self.set_playback_property(PlayProp.Position, 0)

    def on_player_media_loaded(self):
        # As seeked signal is not emitted when media is changed,
        # update position explicitly.
        self.set_playback_property(PlayProp.Position, 0)



import sys

from PyQt6.QtCore import Qt, QRectF, QRect, QSize
from PyQt6.QtGui import (
    QPalette,
    QColor,
    QTextOption,
    QPainter,
    QKeySequence,
    QFont,
    QAction,
    QShortcut,
)
from PyQt6.QtWidgets import (
    QLabel,
    QWidget,
    QVBoxLayout,
    QSizeGrip,
    QHBoxLayout,
    QColorDialog,
    QMenu,
    QFontDialog,
    QSpacerItem,
)

from feeluown.i18n import t
from feeluown.gui.helpers import esc_hide_widget, resize_font, elided_text
from feeluown.player import LyricLine


IS_MACOS = sys.platform == "darwin"
IS_WINDOWS = sys.platform == "win32"


def set_bg_color(palette, color):
    palette.setColor(QPalette.ColorGroup.Active, QPalette.ColorRole.Window, color)
    palette.setColor(QPalette.ColorGroup.Active, QPalette.ColorRole.Base, color)
    palette.setColor(QPalette.ColorGroup.Inactive, QPalette.ColorRole.Window, color)
    palette.setColor(QPalette.ColorGroup.Inactive, QPalette.ColorRole.Base, color)


def set_fg_color(palette, color):
    palette.setColor(QPalette.ColorGroup.Active, QPalette.ColorRole.WindowText, color)
    palette.setColor(QPalette.ColorGroup.Active, QPalette.ColorRole.Text, color)
    palette.setColor(QPalette.ColorGroup.Inactive, QPalette.ColorRole.WindowText, color)
    palette.setColor(QPalette.ColorGroup.Inactive, QPalette.ColorRole.Text, color)


Tooltip = """
* Right-click can pop up the settings menu
* Ctrl+= or Ctrl++ can increase font size
* Ctrl+- can decrease font size
* Mouse forward/back buttons can play the previous/next track
* ESC key can close this lyrics window
"""


class SizeGrip(QSizeGrip):
    """
    On windows, when the user drags the size grip, the lyric window is
    resized and the lyric window also moves. This is not the expected.
    So override the mouse event to fix this issue.

    Check https://github.com/feeluown/FeelUOwn/issues/752 for more details.
    """

    def mousePressEvent(self, e):
        super().mousePressEvent(e)
        if IS_WINDOWS:
            e.accept()

    def mouseMoveEvent(self, e):
        super().mouseMoveEvent(e)
        if IS_WINDOWS:
            e.accept()

    def mouseReleaseEvent(self, e):
        super().mouseReleaseEvent(e)
        if IS_WINDOWS:
            e.accept()

    def paintEvent(self, _):
        """
        On windows, it draws a icon on the corner. On other platforms, it does not.
        So let LyricWindow draw the icon for the SizeGrip.
        """
        if IS_WINDOWS:
            return


class LyricWindow(QWidget):
    """LyricWindow is a transparent container which contains a real lyric window.

    LyricWindow acts as a transparent container, so the inner window can has
    semi-transparent background. It is also responsible for handling the
    window flags and geometry. It also provides a few APIs for communicating
    with other widgets
    """

    def __init__(self, app):
        super().__init__(parent=None)
        self._app = app

        if IS_MACOS:
            # On macOS, Qt.Tooltip widget can't accept focus and it will hide
            # when the application window is actiavted. Qt.Tool widget can't
            # keep staying on top. Neither of them work well on macOS.
            flags = (
                Qt.WindowType.WindowStaysOnTopHint | Qt.WindowType.FramelessWindowHint
            )
        else:
            # TODO: use proper flags on other platforms, see #413 for more details.
            # User can customize the flags in the .fuorc or searchbox, like
            #    app.ui.lyric_windows.setWindowFlags(Qt.xx | Qt.yy)
            flags = (
                Qt.WindowType.WindowStaysOnTopHint
                | Qt.WindowType.FramelessWindowHint
                | Qt.WindowType.Tool
            )
        self.setWindowFlags(flags)
        self.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground)
        self.setToolTip(Tooltip)

        self._inner = InnerLyricWindow(self._app, self)
        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)
        self._layout.addWidget(self._inner)

        self._old_pos = None

        esc_hide_widget(self)

    def mousePressEvent(self, e):
        self._old_pos = e.globalPosition()

    def mouseMoveEvent(self, e):
        # NOTE: e.button() == Qt.MouseButton.LeftButton don't work on Windows
        # on Windows, even I drag with LeftButton, the e.button() return 0,
        # which means no button
        if self._old_pos is not None:
            delta = e.globalPosition() - self._old_pos
            self.move(int(self.x() + delta.x()), int(self.y() + delta.y()))
            self._old_pos = e.globalPosition()

    def mouseReleaseEvent(self, e):
        if not self.rect().contains(e.position().toPoint()):
            return
        if e.button() == Qt.MouseButton.BackButton:
            self._app.playlist.previous()
        elif e.button() == Qt.MouseButton.ForwardButton:
            self._app.playlist.next()

    def dump_state(self):
        inner = self._inner
        p = inner.palette()
        geo = self.geometry()
        return {
            "geometry": (geo.x(), geo.y(), geo.width(), geo.height()),
            "font": inner.font().toString(),
            "bg": p.color(QPalette.ColorGroup.Active, QPalette.ColorRole.Window).name(
                QColor.NameFormat.HexArgb
            ),
            "fg": p.color(QPalette.ColorGroup.Active, QPalette.ColorRole.Text).name(
                QColor.NameFormat.HexArgb
            ),
        }

    def apply_state(self, state):
        if not state:
            return

        inner = self._inner

        geo = state.get("geometry")
        if geo:
            self.resize(geo[2], geo[3])
            self.setGeometry(*geo)
        font = inner.font()
        font.fromString(state["font"])
        inner.setFont(font)
        palette = inner.palette()
        set_bg_color(palette, QColor(state["bg"]))
        set_fg_color(palette, QColor(state["fg"]))
        inner.setPalette(palette)

    def sizeHint(self):
        return QSize(500, 60)

    def resizeEvent(self, e):
        return super().resizeEvent(e)


class SentenceLabel(QLabel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.setAlignment(
            Qt.AlignmentFlag.AlignBaseline
            | Qt.AlignmentFlag.AlignVCenter
            | Qt.AlignmentFlag.AlignHCenter
        )
        self.setWordWrap(False)


class LineLabel(QWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.v_spacing = 10
        self.show_trans = True  # Show translated lyric or not.
        self.label = SentenceLabel(parent=self)
        self.trans_label = SentenceLabel(parent=self)
        self.spacer = QSpacerItem(0, 0)

        self._layout = QVBoxLayout(self)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)

        self._layout.addSpacing(self.v_spacing // 2)
        self._layout.addWidget(self.label)
        self._layout.addSpacerItem(self.spacer)
        self._layout.addWidget(self.trans_label)
        self._layout.addSpacing(self.v_spacing // 2)

        line = LyricLine("...", "...", False)
        self.set_line(line)
        # The default size(calculated by Qt) may be different from the size
        # calculated by line_sizehint. Remember to specify size at very first,
        # otherwise it may show with the default size(calculated by Qt).
        self.resize(self.line_sizehint(line))

    def toggle_show_trans(self):
        self.show_trans = not self.show_trans
        if self.show_trans:
            self.trans_label.show()
            self.spacer.changeSize(0, self.v_spacing // 3)
        else:
            self.trans_label.hide()
            self.spacer.changeSize(0, 0)

    def set_line(self, line: LyricLine):
        self.label.setText(elided_text(line.origin, self.width(), self.font()))
        if self.show_trans and line.has_trans:
            self.trans_label.show()
            self.trans_label.setText(elided_text(line.trans, self.width(), self.font()))
        else:
            self.trans_label.hide()

    def line_sizehint(self, line: LyricLine):
        """Proper size to show the line."""
        rect = self.label.fontMetrics().boundingRect(line.origin)
        height = rect.height()
        if self.show_trans and line.has_trans:
            height = height * 2
        # Sometimes width is not enough for text, so add buffer.
        h_buffer = rect.height()
        # Add some padding for vertical so that it looks more beautiful.
        v_buffer = rect.height() // 4
        height += self.v_spacing + v_buffer + self.spacer.geometry().height()
        return QSize(rect.width() + h_buffer, height)

    def setFont(self, font):
        super().setFont(font)

        self.label.setFont(font)
        font2 = QFont(font)
        font2.setBold(False)
        if font.pointSize() != 0:
            resize_font(font2, -4)
        else:
            resize_font(font2, -8)
        self.trans_label.setFont(font2)

    def setPalette(self, palette):
        super().setPalette(palette)
        self.label.setPalette(palette)
        self.trans_label.setPalette(palette)


class InnerLyricWindow(QWidget):
    """
    This window is responsible for rendering one line of a lyric.
    This window need not to know which song is playing, or if
    the song is changed.
    """

    def __init__(self, app, parent=None):
        super().__init__(parent=parent)
        self._app = app

        self._border_radius = 0
        # When _auto_resize is True,
        # the window size adapts to the lyric sentence width.
        self._auto_resize = True
        self._size_grip = SizeGrip(self)
        self.line_label = LineLabel(self)

        self._app.live_lyric.line_changed.connect(self.set_line)
        self._app.live_lyric.lyrics_changed.connect(self.on_lyrics_changed)
        QShortcut(QKeySequence.StandardKey.ZoomIn, self).activated.connect(self.zoomin)
        QShortcut(QKeySequence.StandardKey.ZoomOut, self).activated.connect(
            self.zoomout
        )
        QShortcut(QKeySequence("Ctrl+="), self).activated.connect(self.zoomin)

        self._layout = QHBoxLayout(self)
        self.setup_ui()

    def setup_ui(self):
        if self._auto_resize:
            self._size_grip.hide()
        self.on_font_size_changed()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)
        self._layout.addWidget(self.line_label)
        self._layout.addWidget(self._size_grip)
        self._layout.setAlignment(self._size_grip, Qt.AlignmentFlag.AlignBottom)

    def set_line(self, line: LyricLine):
        # Ignore updating when the window is invisible.
        if not self.isVisible():
            return
        size = self.line_label.line_sizehint(line)
        self.line_label.resize(size)
        if self._auto_resize:
            self_size = QSize(size.width(), size.height())
            self.resize(self_size)
            self.parent().resize(self_size)  # type: ignore
            self.parent().updateGeometry()  # type: ignore
        self.line_label.set_line(line)

    def on_lyrics_changed(self, lyric, *_):
        if lyric is None:
            self.set_line(LyricLine(t("lyric-not-available"), "", False))

    def zoomin(self):
        font = self.font()
        resize_font(font, +1)
        self.setFont(font)

    def zoomout(self):
        font = self.font()
        resize_font(font, -1)
        self.setFont(font)

    def on_font_size_changed(self):
        self._border_radius = self.fontMetrics().height() // 3
        width = max(1, self._border_radius * 2)
        self._size_grip.setFixedWidth(width)

    def paintEvent(self, e):
        """Draw shapes to make the size_grip more obvious.

        Note the shapes can't be drawed on the outside container (LyricWindow)
        due to it sets the attribute WA_TranslucentBackground.
        """
        painter = QPainter(self)
        painter.setRenderHint(QPainter.RenderHint.Antialiasing)
        painter.setPen(Qt.PenStyle.NoPen)
        painter.setBrush(self.palette().color(QPalette.ColorRole.Window))
        painter.drawRoundedRect(self.rect(), self._border_radius, self._border_radius)

        # Draw an circle button to indicate that the window can be resized.
        if self._auto_resize:
            return
        painter.save()
        painter.setPen(QColor("white"))
        option = QTextOption()
        option.setAlignment(
            Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignVCenter
        )
        rect = QRect(self.mapToParent(self._size_grip.pos()), self._size_grip.size())
        painter.drawText(QRectF(rect), "●", option)
        painter.restore()

    def setPalette(self, a0: QPalette) -> None:
        super().setPalette(a0)
        self.line_label.setPalette(a0)

    def setFont(self, a0: QFont) -> None:
        super().setFont(a0)
        self.line_label.setFont(a0)
        self.on_font_size_changed()

    def show_color_dialog(self, bg=True):
        def set_color(color):
            palette = self.palette()
            if bg:
                set_bg_color(palette, color)
            else:
                set_fg_color(palette, color)
            # Note that this widget(self) must also set the palette,
            # so the background can work as expected.
            self.setPalette(palette)

        dialog = QColorDialog(self)
        # Set WA_DeleteOnClose so that the dialog can be deleted (from self.children).
        dialog.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)
        if bg:
            color = self.palette().color(QPalette.ColorRole.Window)
        else:
            color = self.palette().color(QPalette.ColorRole.Text)
        dialog.setCurrentColor(color)
        dialog.currentColorChanged.connect(set_color)
        dialog.colorSelected.connect(set_color)
        dialog.setOption(QColorDialog.ColorDialogOption.ShowAlphaChannel, True)
        # On KDE(with Xorg), if the dialog is in modal state,
        # the window is dimming.
        if sys.platform == "linux":
            dialog.show()
        else:
            dialog.open()

    def show_font_dialog(self):
        dialog = QFontDialog(self.font(), self)
        # Set WA_DeleteOnClose so that the dialog can be deleted (from self.children).
        dialog.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)
        dialog.currentFontChanged.connect(self.setFont)
        dialog.fontSelected.connect(self.setFont)
        dialog.open()

    def toggle_auto_resize(self):
        self._auto_resize = not self._auto_resize
        if self._auto_resize:
            self._size_grip.hide()
        else:
            self._size_grip.show()

    def contextMenuEvent(self, e):
        menu = QMenu()
        bg_color_action = QAction(t("lyric-background-color"), menu)
        fg_color_action = QAction(t("lyric-text-color"), menu)
        font_action = QAction(t("lyric-font"), menu)
        toggle_trans_action = QAction(t("lyric-show-bilingual"), menu)
        toggle_trans_action.setCheckable(True)
        toggle_trans_action.setChecked(self.line_label.show_trans)
        toggle_fiexed_size_action = QAction(t("lyric-window-auto-resize"), menu)
        toggle_fiexed_size_action.setCheckable(True)
        toggle_fiexed_size_action.setChecked(self._auto_resize)
        menu.addAction(bg_color_action)
        menu.addAction(fg_color_action)
        menu.addSeparator()
        menu.addAction(font_action)
        menu.addSeparator()
        menu.addAction(toggle_trans_action)
        menu.addAction(toggle_fiexed_size_action)

        bg_color_action.triggered.connect(lambda: self.show_color_dialog(bg=True))
        fg_color_action.triggered.connect(lambda: self.show_color_dialog(bg=False))
        font_action.triggered.connect(self.show_font_dialog)
        toggle_trans_action.triggered.connect(self.line_label.toggle_show_trans)
        toggle_fiexed_size_action.triggered.connect(self.toggle_auto_resize)

        menu.exec(e.globalPos())

    def showEvent(self, e) -> None:
        self.set_line(self._app.live_lyric.current_line)
        return super().showEvent(e)



from typing import TYPE_CHECKING

from PyQt6.QtCore import Qt, QSize
from PyQt6.QtGui import QFontMetrics, QPalette, QColor, QPainter
from PyQt6.QtWidgets import QHBoxLayout, QStackedLayout, QWidget

from feeluown.gui.components.player_playlist import PlayerPlaylistView
from feeluown.gui.helpers import (
    fetch_cover_wrapper,
    palette_set_bg_color,
    random_solarized_color,
)
from feeluown.gui.widgets import MVButton
from feeluown.gui.widgets.comment_list import CommentListView, CommentListModel
from feeluown.gui.widgets.cover_label import CoverLabelV2
from feeluown.gui.widgets.lyric import LyricView
from feeluown.gui.widgets.song_minicard_list import (
    SongMiniCardListDelegate,
    SongMiniCardListView,
    SongMiniCardListModel,
)
from feeluown.i18n import t
from feeluown.library import SupportsSongHotComments, SupportsSongSimilar
from feeluown.utils.aio import run_fn, run_afn
from feeluown.utils.reader import create_reader

if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp


class RefreshOnSongChangedMixin:
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._need_refresh = True
        self._app.playlist.song_changed.connect(
            lambda song: run_afn(self.on_song_changed, song), aioqueue=True, weak=False
        )

    async def on_song_changed(self, _):
        if not self.isVisible():
            self._need_refresh = True
            return
        self.run_refresh_task()

    def showEvent(self, e):
        super().showEvent(e)
        if self._need_refresh:
            self.run_refresh_task()

    def run_refresh_task(self):
        self._app.task_mgr.run_afn_preemptive(self.refresh)
        self._need_refresh = False


class MVWrapper(QWidget):
    def __init__(self, parent):
        super().__init__(parent=parent)

        palette = self.palette()
        color = QColor("black")
        color.setAlpha(150)
        palette_set_bg_color(palette, color)
        palette.setColor(QPalette.ColorRole.Text, QColor("white"))
        palette.setColor(QPalette.ColorRole.WindowText, QColor("white"))
        palette.setColor(QPalette.ColorRole.ButtonText, QColor("white"))
        self.setPalette(palette)

        self._layout = QHBoxLayout(self)
        self.mv_btn = MVButton(length=20, parent=self)
        self.mv_btn.setToolTip(t("track-movie-play-tooltip"))

        self.mv_btn.setPalette(palette)
        self._layout.addWidget(self.mv_btn)
        self._layout.setContentsMargins(0, 0, 0, 0)
        self.setFixedSize(40, 25)

    def paintEvent(self, _):
        painter = QPainter(self)
        painter.setPen(Qt.PenStyle.NoPen)
        painter.setBrush(self.palette().color(QPalette.ColorRole.Window))
        painter.drawRoundedRect(self.rect(), 3, 3)


class NowplayingArtwork(QWidget):
    def __init__(self, app: "GuiApp", parent=None):
        super().__init__(parent=parent)

        self._app = app
        self._inner = CoverLabelV2(app, self)
        self._mv_wrapper = MVWrapper(parent=self)
        self.mv_btn = self._mv_wrapper.mv_btn

        self._app.player.metadata_changed.connect(
            self.on_metadata_changed, aioqueue=True
        )
        self._app.playlist.song_mv_changed.connect(
            self.on_current_song_mv_changed, aioqueue=True
        )

        self._stacked_layout = QStackedLayout(self)
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)
        self._stacked_layout.addWidget(self._inner)
        self.layout().setAlignment(self._inner, Qt.AlignmentFlag.AlignVCenter)

        self._mv_wrapper.raise_()

    def set_body(self, widget=None):
        if widget is None:
            widget = self._inner
        if widget == self._inner:
            self._stacked_layout.setCurrentWidget(self._inner)
            self._mv_wrapper.raise_()
        else:
            self._stacked_layout.addWidget(widget)
            self._stacked_layout.setCurrentWidget(widget)
            self._stacked_layout.setAlignment(widget, Qt.AlignmentFlag.AlignVCenter)
            self._mv_wrapper.hide()

    def on_current_song_mv_changed(self, _, mv):
        if mv is not None:
            self._mv_wrapper.show()
        else:
            self._mv_wrapper.hide()

    def on_metadata_changed(self, metadata):
        metadata = metadata or {}
        released = metadata.get("released", "")
        if released:
            self.setToolTip(t("track-album-release-date", releaseDate=released))
        else:
            self.setToolTip("")
        # Set song artwork.
        artwork = metadata.get("artwork", "")
        source = metadata.get("source", "")
        artwork_uid = metadata.get("uri", artwork)
        if artwork:
            run_afn(self._inner.show_cover_with_source, artwork, source, artwork_uid)
        else:
            self._inner.show_img(None)

    def resizeEvent(self, a0) -> None:
        super().resizeEvent(a0)
        self._mv_wrapper.move(
            self.width() - self._mv_wrapper.width() - 10,
            self.height() - self._mv_wrapper.height() - 10,
        )


class NowplayingLyricView(LyricView):
    """
    Let user zoom in/out.
    """

    def __init__(self, app: "GuiApp", parent=None):
        super().__init__(parent=parent)
        self._app = app
        self._app.live_lyric.lyrics_changed.connect(
            self.on_lyric_changed, aioqueue=True
        )
        self._app.live_lyric.line_changed.connect(self.on_line_changed, weak=True)

        font = self.font()
        font.setPixelSize(17)
        self.setFont(font)

        self._alignment = Qt.AlignmentFlag.AlignCenter
        self._highlight_font_size = 25
        self._item_spacing = 20

        palette = self.palette()
        palette.setColor(QPalette.ColorRole.Highlight, Qt.GlobalColor.transparent)
        palette.setColor(QPalette.ColorRole.HighlightedText, random_solarized_color())
        self.setPalette(palette)

    def _create_item(self, line):
        item = super()._create_item(line)
        rect = QFontMetrics(item.font()).boundingRect(line)
        size = QSize(rect.width(), rect.height() + self._item_spacing)
        item.setData(Qt.ItemDataRole.UserRole, (line, size))
        item.setSizeHint(size)
        return item

    def on_item_changed(self, current, previous):
        super().on_item_changed(current, previous)

        if current:
            line, size_hint = current.data(Qt.ItemDataRole.UserRole)
            rect = QFontMetrics(current.font()).boundingRect(line)
            current.setSizeHint(
                QSize(
                    size_hint.width(),
                    rect.height() + self._item_spacing,
                )
            )

    def reset_item(self, item):
        super().reset_item(item)
        if item:
            item.setSizeHint(item.data(Qt.ItemDataRole.UserRole)[1])


class NowplayingCommentListView(RefreshOnSongChangedMixin, CommentListView):
    def __init__(self, app: "GuiApp", parent=None):
        self._app = app
        super().__init__(
            parent=parent,
            no_scroll_v=False,
            delegate_options={"quoted_bg_color_role": QPalette.ColorRole.Base},
        )

    async def refresh(self):
        song = self._app.playlist.current_song
        reader = create_reader([])
        if song is not None:
            provider = self._app.library.get(song.source)
            if isinstance(provider, SupportsSongHotComments):
                comments = await run_fn(provider.song_list_hot_comments, song)
                reader = create_reader(comments)
        self.setModel(CommentListModel(reader))  # type: ignore


class NowplayingSimilarSongsView(RefreshOnSongChangedMixin, SongMiniCardListView):
    def __init__(self, app: "GuiApp", parent=None):
        self._app = app
        super().__init__(parent=parent, no_scroll_v=False)
        self.setItemDelegate(
            SongMiniCardListDelegate(
                self,
                card_min_width=200,
                card_height=40,
                card_padding=(5 + SongMiniCardListDelegate.img_padding, 5, 0, 5),
                card_right_spacing=10,
                hover_color_role=QPalette.ColorRole.Base,
            )
        )

    async def refresh(self):
        song = self._app.playlist.current_song
        reader = create_reader([])
        if song is not None:
            provider = self._app.library.get(song.source)
            if isinstance(provider, SupportsSongSimilar):
                songs = await run_fn(provider.song_list_similar, song)
                reader = create_reader(songs)
        self.setModel(SongMiniCardListModel(reader, fetch_cover_wrapper(self._app)))


class NowplayingPlayerPlaylistView(PlayerPlaylistView):
    def __init__(self, app, parent=None, **kwargs):
        kwargs.setdefault("no_scroll_v", False)
        kwargs.setdefault("row_height", 60)
        super().__init__(app, parent=parent, **kwargs)

        delegate = SongMiniCardListDelegate(
            self,
            card_min_width=200,
            card_height=40,
            card_padding=(5 + SongMiniCardListDelegate.img_padding, 5, 0, 5),
            card_right_spacing=10,
            hover_color_role=QPalette.ColorRole.Base,
        )
        self.setItemDelegate(delegate)



import random

from feeluown.collection import Collection
from feeluown.library import (
    BriefSongModel,
    ModelType,
    reverse,
    BriefAlbumModel,
    BriefArtistModel,
)


def song2line(song: BriefSongModel):
    line = reverse(song, as_line=True)
    parts = line.split("#", 1)
    if len(parts) >= 2:
        return parts[1]
    return None


def artist2line(artist: BriefArtistModel):
    line = reverse(artist, as_line=True)
    parts = line.split("#", 1)
    if len(parts) >= 2:
        return parts[1]
    return None


def album2line(album: BriefAlbumModel):
    line = reverse(album, as_line=True)
    parts = line.split("#", 1)
    if len(parts) >= 2:
        return parts[1]
    return None


async def generate_prompt_for_library(library: Collection):
    """Generate prompt based on the user's music library."""
    songs = []
    artists = []
    albums = []
    for model in library.models:
        if ModelType(model.meta.model_type) is ModelType.song:
            songs.append(model)
        elif ModelType(model.meta.model_type) is ModelType.artist:
            artists.append(model)
        elif ModelType(model.meta.model_type) is ModelType.album:
            albums.append(model)
    random.shuffle(songs)
    random.shuffle(artists)
    random.shuffle(albums)
    random_songs = songs[:10]
    random_artists = artists[:10]
    random_albums = albums[:10]

    song_lines = []
    for song in random_songs:
        line = song2line(song)
        if line:
            song_lines.append("    " + line)
    songs_text = "\n".join(song_lines)

    artist_lines = []
    for artist in random_artists:
        line = artist2line(artist)
        if line:
            artist_lines.append("    " + line)
    artists_text = "\n".join(artist_lines)

    album_lines = []
    for album in random_albums:
        line = album2line(album)
        if line:
            album_lines.append("    " + line)
    albums_text = "\n".join(album_lines)

    prompt_text = f"""用户的音乐库包含以下内容：
歌曲（总数：{len(songs)}，随机抽取 {len(song_lines)} 首）：
{songs_text}
歌手（总数：{len(artists)}，随机抽取 {len(artist_lines)} 名）：
{artists_text}
专辑（总数：{len(albums)}，随机抽取 {len(album_lines)} 个）：
{albums_text}
"""
    return prompt_text



from .textbtn import TextButton  # noqa
from .login import CookiesLoginDialog  # noqa
from .selfpaint_btn import (  # noqa
    SelfPaintAbstractSquareButton,
    RecentlyPlayedButton,
    HomeButton,
    LeftArrowButton,
    RightArrowButton,
    SearchSwitchButton,
    SettingsButton,
    PlusButton,
    TriagleButton,
    DiscoveryButton,
    SelfPaintAbstractIconTextButton,
    CalendarButton,
    RankButton,
    StarButton,
    PlayPauseButton,
    PlayNextButton,
    PlayPreviousButton,
    PlayButton,
    MVButton,
    VolumeButton,
    HotButton,
    EmojiButton,
    AIButton,
)



import asyncio
import logging

from feeluown.server.pubsub import (
    Gateway as PubsubGateway,
    LiveLyricPublisher,
)
from feeluown.server.pubsub.publishers import SignalPublisher
from feeluown.server import FuoServer, ProtocolType
from feeluown.nowplaying import run_nowplaying_server
from .app import App

logger = logging.getLogger(__name__)


class ServerApp(App):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.rpc_server = FuoServer(self, ProtocolType.rpc)
        self.pubsub_server = FuoServer(self, ProtocolType.pubsub)
        self.pubsub_gateway = PubsubGateway()
        self._ll_publisher = LiveLyricPublisher(self.pubsub_gateway)
        self._signal_publish = SignalPublisher(self.pubsub_gateway)

    def initialize(self):
        super().initialize()
        self.live_lyric.sentence_changed.connect(self._ll_publisher.publish)

        signals = [
            (self.player.metadata_changed, 'player.metadata_changed'),
            (self.player.seeked, 'player.seeked'),
            (self.player.state_changed, 'player.state_changed'),
            (self.player.duration_changed, 'player.duration_changed'),
            (self.live_lyric.sentence_changed, 'live_lyric.sentence_changed'),
        ]
        for signal, name in signals:
            self.pubsub_gateway.add_topic(name)
            signal.connect(self._signal_publish.on_emitted(name),
                           weak=False,
                           aioqueue=True)

    def run(self):
        super().run()
        listen_addr = self.get_listen_addr()

        asyncio.create_task(self.rpc_server.run(listen_addr, self.config.RPC_PORT))
        asyncio.create_task(
            self.pubsub_server.run(
                listen_addr,
                self.config.PUBSUB_PORT,
            ))
        if self.config.ENABLE_WEB_SERVER:
            try:
                from feeluown.webserver import run_web_server
            except ImportError as e:
                logger.error(f"can't enable webserver, err: {e}")
            else:
                asyncio.create_task(run_web_server(listen_addr, self.config.WEB_PORT))
        if self.config.ENABLE_MCP_SERVER:
            try:
                from feeluown.mcpserver import run_mcp_server
                mcp_task = run_mcp_server(
                    listen_addr,
                    self.config.MCP_PORT,
                    debug=bool(self.config.DEBUG),
                )
            except ImportError as e:
                logger.error(f"can't enable mcp server, err: {e}")
            else:
                asyncio.create_task(mcp_task)
        asyncio.create_task(run_nowplaying_server(self))



"""
Playlist Management
~~~~~~~~~~~~~
"""

from feeluown.gui.widgets.playlists import PlaylistsModel

from feeluown.library import PlaylistModel


class PlaylistUiItem(PlaylistModel):
    """
Based on current experience, the most basic operations related to playlists are:

    * Creating, deleting
    * Adding, removing songs
    * Renaming
    * Clicking to display this playlist

    These operations have consistent semantics across platforms,
    so PlaylistUiItem does not provide signals like clicked for now.
    """


class PlaylistUiManager:
    def __init__(self, app):
        self._app = app
        self.model = PlaylistsModel(app)

    def add(self, playlist, is_fav=False):
        self.model.add(playlist, is_fav=is_fav)

    def clear(self):
        self.model.clear()



"""
Testing the performance difference between accessing `obj.attr` directly and using `await async_run(lambda: obj.attr)`.

Test results: accessing `obj.attr` 100,000 times takes milliseconds, whereas the latter takes seconds (around 10 seconds).
"""


import asyncio
import time


class Obj:
    def __init__(self):
        self.attr = 1

    @property
    def prop(self):
        return 1


async def f():
    loop = asyncio.get_event_loop()
    obj = Obj()

    t = time.time()
    t_cpu = time.process_time()

    for _ in range(0, 100000):
        await loop.run_in_executor(None, lambda: obj.attr)

    print(time.time() - t )
    print(time.process_time() - t_cpu)



if __name__ == '__main__':
    loop = asyncio.get_event_loop()
    loop.run_until_complete(f())



import logging
from typing import Optional

from feeluown.server.data_structure import Request, Response
from feeluown.server.session import SessionLike
from feeluown.serializers import serialize
from .base import cmd_handler_mapping, AbstractHandler
from .cmd import Cmd
from .excs import HandlerException

logger = logging.getLogger(__name__)

V1_SUPPORTTED_CMDS = [
    'status', 'play', 'pause', 'resume', 'toggle',
    'stop', 'next', 'previous', 'search', 'show',
    'list', 'clear', 'remove', 'add', 'exec',
]


async def handle_request(
        req: Request,
        app,
        session: Optional[SessionLike] = None,
) -> Response:
    """
    :type req: feeluown.server.rpc.Request
    """
    cmd = Cmd(req.cmd, *req.cmd_args, options=req.cmd_options)

    if session is not None and \
       session.options.rpc_version == '1.0' and \
       cmd.action not in V1_SUPPORTTED_CMDS:
        handler_cls = None
    else:
        handler_cls = cmd_handler_mapping.get(cmd.action)

    ok, body = False, ''
    if handler_cls is None:
        ok, body = False, f"handler for cmd:{req.cmd} not found"
    else:
        try:
            handler: AbstractHandler = handler_cls(app=app, session=session)
            if handler.support_aio_handle is True:
                rv = await handler.a_handle(cmd)
            else:
                # FIXME: handlers do blocking io should implement a_handle method.
                rv = handler.handle(cmd)
        except HandlerException as e:
            ok, body = False, str(e)
        except Exception as e:  # pylint: disable=broad-except
            logger.exception(f'handle cmd({cmd}) error')
            ok, body = False, f'internal server error\n{str(e)}'
        else:
            rv = rv if rv is not None else ''
            ok, body = True, rv
    fmt = req.options.get('format', None)
    fmt = fmt or 'plain'
    msg = serialize(fmt, body, brief=False)
    return Response(ok=ok, text=msg, req=req)


# pylint: disable=wrong-import-position, cyclic-import
from .help import HelpHandler  # noqa
from .status import StatusHandler  # noqa
from .playlist import PlaylistHandler  # noqa
from .player import PlayerHandler  # noqa
from .search import SearchHandler  # noqa
from .show import ShowHandler  # noqa
from .exec_ import ExecHandler  # noqa
from .sub import SubHandler  # noqa
from .set_ import SetHandler  # noqa
try:
    from .jsonrpc_ import JsonRPCHandler  # noqa
except ImportError as e:
    logger.info(f'jsonrpc handler is not available: {e}')



import os
from pathlib import Path
import sys
import logging
from decimal import Decimal
from datetime import date, datetime
from importlib import resources
from threading import RLock
from collections import defaultdict

import langcodes
from fluent.runtime import FluentBundle, FluentLocalization, FluentResourceLoader

import feeluown.i18n

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


_L10N_BUNDLE: dict[tuple[str, ...], FluentLocalization] = {}
_L10N_BUNDLE_LOCK: dict[tuple[str, ...], RLock] = defaultdict(RLock)


def rfc1766_langcode() -> str:
    """
    Returns a RFC 1766 language code, for current user preference.
    """

    import locale

    match sys.platform:
        case "win32":
            from .windows import user_default_locale

            lang = user_default_locale()
        case _:
            lang, _ = locale.getlocale(locale.LC_CTYPE)
            if lang == "C":
                lang = "en_US"

    return lang


# BCP-47 language code
OVERRIDE_LOCALE = os.environ.get("FEELUOWN_LOCALE", None)
# Default locale
_DEFAULT_LOCALE = OVERRIDE_LOCALE if OVERRIDE_LOCALE is not None else rfc1766_langcode()


def t(
    msg_id: str,
    locale: str = None,
    **kwargs: str | int | float | Decimal | date | datetime | object,
) -> str:
    """
    :param msg_id: Message ID inside fluent translation files.
    :param locale: Optional BCP-47 language code
    :param kwargs: Any object that implements the `__str__`

    If `locale` was left `None`, this will use the _DEFAULT_LOCALE.

    To format DATETIME() correctly, you must pass a date/datetime object.
    """
    if locale is None:
        locale = _DEFAULT_LOCALE

    for k, v in kwargs.items():
        # Special types
        if not isinstance(
            v,
            (
                str,
                float,
                int,
                date,
                datetime,
                Decimal,
            ),
        ):
            kwargs[k] = str(v)

    return l10n_bundle(locale).format_value(msg_id, kwargs)


def l10n_bundle(locale: str | None = None) -> FluentLocalization:
    """
    You could call this method multiple times

    **NOTE:** `locale.getlocale` from standard library is unreliable.

    :param locale: BCP-47 language code.
    :return: l10n localization bundle
    """
    if locale is None:
        if OVERRIDE_LOCALE is None:
            locale = rfc1766_langcode()
        else:
            locale = OVERRIDE_LOCALE

    return load_l10n_resource(locales=[locale])


def load_l10n_resource(
    locales: list[str],
    skip_fallback: bool = False,
    resource_ids: list[str] = None,
) -> FluentLocalization:
    if resource_ids is None:
        resource_ids = DEFAULT_RESOURCE_IDS

    with resources.as_file(
        resources.files(feeluown.i18n) / "assets",
    ) as current_dir:
        supported = [lang for lang in os.listdir(current_dir)]
        # The 'str' typing hint of `roots` is incorrect
        res_loader = FluentResourceLoader(roots=[current_dir / "{locale}"])

        matched_locales = []
        for locale in locales:
            matched_best = langcodes.closest_supported_match(
                desired_language=locale,
                supported_languages=supported,
            )
            if matched_best is not None:
                matched_locales.append(matched_best)

        locales_to_load = matched_locales
        if not skip_fallback:
            locales_to_load += ["en-US", "zh-CN"]

        cache_key = (tuple(locales_to_load), tuple(resource_ids))
        with _L10N_BUNDLE_LOCK[cache_key]:
            if cache_key in _L10N_BUNDLE:
                return _L10N_BUNDLE[cache_key]

            logger.info(f"Loading locale for: {locales_to_load}")

            # resources are loaded immediately,
            # so current_dir can be cleaned safely.
            bundle = FluentLocalization(
                # add en-US, zh-CN for fallback
                locales=locales_to_load,
                resource_ids=resource_ids,
                resource_loader=res_loader,
            )

            _L10N_BUNDLE[cache_key] = bundle

        return bundle


def human_readable_number(n: int, locale: str = None) -> str:
    """
    Compact number formatter

    :param n: the number
    :param locale: BCP-47 language code,
    """

    if locale is None:
        locale = _DEFAULT_LOCALE

    if locale.startswith("zh"):
        levels = [
            (100000000, "亿"),
            (10000, "万"),
        ]
    else:
        levels = [
            (1_000_000_000, "B"),
            (1_000_000, "M"),
            (1_000, "K"),
        ]

    for value, unit in levels:
        if n > value:
            first, second = n // value, (n % value) // (value // 10)
            return f"{first}.{second}{unit}"
    return str(n)


DEFAULT_RESOURCE_IDS = ["app.ftl", "argparser.ftl", "config.ftl"]

if __name__ == "__main__":
    for res_id in DEFAULT_RESOURCE_IDS:
        resource_ids = [res_id]
        print(f"""{res_id}:
-----------------------------""")
        l10n_zh = next(
            load_l10n_resource(
                locales=["zh-CN"],
                skip_fallback=True,
                resource_ids=resource_ids,
            )._bundles()
        )
        total_term_len = len(l10n_zh._terms)
        total_msg_len = len(l10n_zh._messages)
        for locale in os.listdir(Path(__file__).parent / "assets"):
            if locale == "zh-CN":
                continue

            try:
                bundle: FluentBundle = next(
                    load_l10n_resource(
                        locales=[locale],
                        skip_fallback=True,
                        resource_ids=resource_ids,
                    )._bundles()
                )
            except StopIteration:
                continue
            print(f"{locale}")

            if total_term_len:
                print(
                    f"Terms: {len(bundle._terms)}/{total_term_len}"
                    f" ({100 * len(bundle._terms) / total_term_len:.1f}%)"
                )
            if total_msg_len:
                print(
                    f"Messages: {len(bundle._messages)}/{total_msg_len}"
                    f" ({100 * len(bundle._messages) / total_msg_len:.1f}%)"
                )

            print()



"""
Table of contents view for one collection
"""

from PyQt6.QtCore import (Qt, QAbstractListModel, QModelIndex, QSize,
                          QRectF, QRect, QPoint, )
from PyQt6.QtGui import (QPainter, QPalette, QPen, QTextOption)
from PyQt6.QtWidgets import QListView, QStyledItemDelegate, QStyle


class CollectionTOCModel(QAbstractListModel):
    def __init__(self, parent=None):
        super().__init__(parent)

        self.items = list(['理性与感性', '单身日志（Live）', '叶慧美', '晴日共剪窗'])

    def rowCount(self, _=QModelIndex()):
        return len(self.items)

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        row = index.row()
        if role == Qt.ItemDataRole.DisplayRole:
            return self.items[row]
        return None


class CollectionTOCDelegate(QStyledItemDelegate):
    def __init__(self, parent=None):
        super().__init__(parent=parent)

    def paint(self, painter, option, index):
        # refer to pixelator.py
        if option.state & QStyle.StateFlag.State_Selected:
            painter.fillRect(option.rect, option.palette.highlight())

        painter.save()
        painter.setRenderHint(QPainter.Antialiasing)

        text_pen = QPen(option.palette.color(QPalette.Text))
        hl_text_pen = QPen(option.palette.color(QPalette.HighlightedText))
        if option.state & QStyle.StateFlag.State_Selected:
            painter.setPen(hl_text_pen)
        else:
            painter.setPen(text_pen)

        # draw circle
        topleft = option.rect.topLeft()
        x, y = topleft.x(), topleft.y()
        r = option.rect.height() // 2
        center_x = x + r
        center_y = y + r
        circle_center = QPoint(center_x, center_y)
        # painter.drawEllipse(circle_center, r, r)

        flags = Qt.AlignmentFlag.AlignmentFlag.AlignCenter
        circle_rect = QRect(x, y, 2*r, 2*r)
        font = painter.font()
        font.setPointSize(r//2)
        painter.setFont(font)

        # draw icon
        # if index.row() % 2 != 0:
        #     painter.drawText(circle_rect, flags, '♪')
        # else:
        #     painter.drawText(circle_rect, flags, '♪')

        pen = painter.pen()
        pen.setCapStyle(Qt.RoundCap)
        pen.setWidth(2)
        painter.setPen(pen)
        r = 22
        r_dis = 5
        painter.drawEllipse(circle_center, r, r)
        painter.drawEllipse(circle_center, r - 3 * r_dis, r - 3 * r_dis)
        painter.drawEllipse(circle_center, r - 4 * r_dis, r - 4 * r_dis)

        # outer arc
        arc_r = r - r_dis
        topleft = QPoint(center_x - arc_r, center_y - arc_r)
        bottomright = QPoint(center_x + arc_r, center_y + arc_r)
        start_angle = 16
        span = 60
        painter.drawArc(QRect(topleft, bottomright), 16 * start_angle, 16 * span)
        painter.drawArc(QRect(topleft, bottomright), 16 * (start_angle + 180), 16 * span)

        # inner arc
        inner_arc_r = arc_r - r_dis
        start_angle = 20
        span = 50
        topleft = QPoint(center_x - inner_arc_r, center_y - inner_arc_r)
        bottomright = QPoint(center_x + inner_arc_r, center_y + inner_arc_r)
        painter.drawArc(QRect(topleft, bottomright), 16 * start_angle, 16 * span)
        painter.drawArc(QRect(topleft, bottomright), 16 * (start_angle + 180), 16 * span)

        text = index.data(Qt.ItemDataRole.DisplayRole)
        topleft = option.rect.topLeft()
        topleft = QPoint(topleft.x() + 2*r + 15, topleft.y())
        painter.drawText(QRect(topleft, option.rect.bottomRight()), Qt.AlignmentFlag.AlignVCenter, text)
        painter.restore()

    def sizeHint(self, option, index):
        if index.isValid():
            return QSize(100, 50)
        return super().sizeHint(option, index)


class CollectionTOCView(QListView):
    def __init__(self, parent=None):
        super().__init__(parent=parent)

        delegate = CollectionTOCDelegate(self)
        self.setItemDelegate(delegate)



if __name__ == '__main__':
    from PyQt6.QtWidgets import QApplication

    app = QApplication([])
    view = CollectionTOCView()
    model = CollectionTOCModel(view)
    view.setModel(model)
    view.show()
    app.exec()



"""
The DSL is used for RPC and resource definition. The syntax of the DSL is same
as UNIX shell.

>>> req = parse("search jaychou -s=xx")
>>> req.cmd
'search'
>>> req.cmd_args
['jaychou']
"""

import argparse
import shlex
import itertools
from typing import Optional

from feeluown.argparser import (
    create_fmt_parser,
    add_common_cmds,
    add_server_cmds,
)
from feeluown.server.data_structure import Request
from feeluown.server.excs import FuoSyntaxError


def tokenize(source):
    s = shlex.shlex(source, punctuation_chars=True, posix=True)
    s.whitespace_split = True
    try:
        tokens = list(s)
    except ValueError as e:
        raise FuoSyntaxError(str(e)) from None
    else:
        if s.lineno >= 2:
            raise FuoSyntaxError("source must be only one line")
        return tokens


class ArgumentParserNoExitAndPrint(argparse.ArgumentParser):
    """
    This customized argument parser class is design to handle two scenario
    1. When there is an error, the parser should not exit. So the error method is
       overrided.
    2. When `help` action is executed, the parser should not exit and it should
       not print help message to stdout/stderr either. So the `_print_message` and
       `exit` method are overrided.
    """
    def _print_message(self, message, file=None):  # noqa
        pass

    def exit(self, status=0, message=None):  # noqa
        pass

    def error(self, message):
        raise FuoSyntaxError(message)


def create_dsl_parser():
    # pylint: disable=protected-access
    parser = ArgumentParserNoExitAndPrint(add_help=False)
    subparsers = parser.add_subparsers(
        dest='cmd',
    )
    add_common_cmds(subparsers)
    add_server_cmds(subparsers)
    return parser


class Parser:
    def __init__(self, source):
        self._source = source

    def parse(self) -> Request:
        """Parse the source to a Request object.

        argparse have little public methods, so some protected methods are used.
        """
        # pylint: disable=too-many-locals,protected-access,too-many-branches
        parser: ArgumentParserNoExitAndPrint = create_dsl_parser()
        tokens = tokenize(self._source)

        # Handle io_here token.
        has_heredoc, heredoc_word = False, None
        for i, token in enumerate(tokens.copy()):
            if token == '<<':  # Heredoc document.
                has_heredoc = True
                try:
                    heredoc_word = tokens.pop(i+1)
                except IndexError:
                    raise FuoSyntaxError('no heredoc word') from None
                else:
                    tokens.pop(i)  # Pop token '<<' from tokens.
            elif token in ('<', '<<<'):
                raise FuoSyntaxError('unknown token')

        # Parse the tokens.
        args, remain = parser.parse_known_args(tokens)
        if remain:
            raise FuoSyntaxError(f'unknown tokens {tokens}')

        # Get cmdname from the parse result.
        cmdname = getattr(args, 'cmd')
        subparser = get_subparser(parser, cmdname)
        assert subparser is not None, f'parser for cmd:{cmdname} not found'

        # Get cmd args from the parse result.
        cmd_args = []
        for action in subparser._positionals._group_actions:
            cmd_args.append(getattr(args, action.dest))

        # Get req options from the parse result.
        req_options = {}
        option_names_req = []
        for parser_ in [create_fmt_parser()]:
            for action in parser_._actions:
                name = action.dest
                option_names_req.append(name)
                value = getattr(args, name)
                req_options[name] = value

        # Get cmd options from the parse result.
        cmd_options = {}
        for action in subparser._optionals._group_actions:
            option_name = action.dest
            if option_name == 'help':  # Ignore help action.
                continue
            if option_name not in option_names_req:
                cmd_options[option_name] = getattr(args, option_name)

        return Request(cmdname,
                       cmd_args,
                       cmd_options,
                       req_options,
                       has_heredoc=has_heredoc, heredoc_word=heredoc_word)


def get_subparser(parser, cmdname) -> Optional[argparse.ArgumentParser]:
    # pylint: disable=protected-access
    # Get cmdname from the parse result.
    root_dest = 'cmd'
    subparser = None
    for action in parser._actions:
        if action.dest == root_dest:
            subparser = action._name_parser_map[cmdname]  # type: ignore
            break
    return subparser


def parse(source):
    return Parser(source).parse()


def unparse(request: Request):
    """Generate source code for the request object"""
    # pylint: disable=protected-access,too-many-branches
    parser = create_dsl_parser()
    subparser = get_subparser(parser, request.cmd)
    if subparser is None:
        raise ValueError(f'{request.cmd}: no such cmd')

    cmdline = [request.cmd]

    # Unparse cmd args.
    if request.has_heredoc:
        cmdline.append(f'<<{request.heredoc_word}')
    else:
        cmdline.extend([shlex.quote(each) for each in request.cmd_args])

    # Unparse cmd options.
    for key, value in itertools.chain(
            request.cmd_options.items(), request.options.items()):
        for action in subparser._actions:
            if action.dest == key:
                # The option has a default value.
                if value is None:
                    break

                if isinstance(action, argparse._StoreTrueAction):
                    if value is True:
                        cmdline.append(f'--{key}')
                elif isinstance(action, argparse._AppendAction):
                    for each in value or []:
                        cmdline.append(f'--{key}={shlex.quote(str(each))}')
                else:
                    cmdline.append(f'--{key}={shlex.quote(str(value))}')
                break
        else:
            raise ValueError(f'{key}: no such option')

    cmdtext = ' '.join(cmdline)
    if request.has_heredoc:
        cmdtext += '\n'
        cmdtext += request.cmd_args[0]
        cmdtext += '\n'
        cmdtext += request.heredoc_word
    return cmdtext



from PyQt6.QtCore import QRectF, Qt
from PyQt6.QtWidgets import QSizeGrip
from PyQt6.QtGui import QTextOption, QPainter


class SizeGrip(QSizeGrip):
    def __init__(self, parent=None):
        super().__init__(parent)

    def paintEvent(self, e):
        painter = QPainter(self)
        option = QTextOption()
        option.setAlignment(
            Qt.AlignmentFlag.AlignHCenter | Qt.AlignmentFlag.AlignVCenter
        )
        painter.drawText(QRectF(self.rect()), "●", option)



from feeluown.i18n import t
from feeluown.utils.aio import run_afn
from feeluown.utils.reader import create_reader
from feeluown.gui.page_containers.table import Renderer


async def render(req, **kwargs):
    app = req.ctx["app"]
    right_panel = app.ui.right_panel
    right_panel.set_body(right_panel.table_container)
    run_afn(app.ui.table_container.set_renderer, PlayerPlaylistRenderer())


class PlayerPlaylistRenderer(Renderer):
    async def render(self):
        self.meta_widget.title = t("recently-played")
        self.meta_widget.show()
        songs = self._app.recently_played.list_songs()
        reader = create_reader(songs)
        self.show_songs(reader)
        self.toolbar.hide()



import random

from feeluown.utils import aio
from feeluown.app import App
from feeluown.i18n import t

# (nickname, name), no order.
# Write your Chinese(or other languages) name if you want.
Contributors = [
    ("cosven", "Shaowen Yin"),
    ("felixonmars", ""),
    ("PhotonQuantum", ""),
    ("wuliaotc", ""),
    ("cyliuu", ""),
    ("light4", ""),
    ("hjlarry", ""),
    ("sbwtw", ""),
    ("oryxfea", ""),
    ("poly000", ""),
    ("BruceZhang1993", ""),
    ("cposture", ""),
    ("helinb", ""),
    ("Yexiaoxing", ""),
    ("albertofwb", ""),
    ("catsout", ""),
    ("Torxed", ""),
    ("RealXuChe", ""),
    ("rapiz1", ""),
    ("Heersin", ""),
    ("chen-chao", ""),
    ("keter42", ""),
    ("timgates42", ""),
    ("junhan-z", ""),
    ("CareF", ""),
    ("berberman", ""),
    ("SaneBow", ""),
    ("wsyxbcl", ""),
    ("Thearas", ""),
    ("reaink", ""),
    ("wenLiangcan", ""),
    ("leedagee", ""),
    ("hanchanli", ""),
    ("xssss1", ""),
    ("williamherry", ""),
    ("seiuneko", ""),
    ("emmanuel-ferdman", ""),
    ("c-xk", "Shinka"),
    ("w568w", ""),
    ("mokurin000", "莯凛"),
]


class TipsManager:
    """Display some usage tips at appropriate times."""

    tips = [
        t("tips-osdlyrics"),
        t("tips-search-shortcut", shortcut="Ctrl + F"),
        t("tips-show-more-tips"),
        t("tips-album-original-image"),
        t("tips-track-drag-to-playlist"),
        t("tips-common-tooltip"),
        t("tips-watch-mode"),
    ]

    def __init__(self, app: App):
        self._app = app
        self._app.started.connect(self.show_random_after_5s)

    def show_random(self):
        if random.randint(0, 10) < 5:
            self._app.show_msg(random.choice(self.tips), timeout=2500)
        else:
            contributor = random.choice(Contributors)
            nickname, name = contributor
            if name:
                user = f"{name} (@{nickname})"
            else:
                user = f"@{nickname}"
            msg = t("thanks-contributor", user=user)
            self._app.show_msg(msg, timeout=2500)

    def show_random_after_5s(self, *_):
        async def task():
            await aio.sleep(5)
            self.show_random()

        aio.run_afn(task)



import sys

from PyQt6.QtCore import Qt
from PyQt6.QtWidgets import QScrollArea, QFrame

from feeluown.gui.helpers import BgTransparentMixin


class ScrollArea(QScrollArea, BgTransparentMixin):
    """
    ScrollArea is designed to be used as a container for page_body.
    Be careful when you use it as a container for other widgets.
    """

    def __init__(self, parent=None):
        super().__init__(parent=parent)

        self.setWidgetResizable(True)
        self.setFrameShape(QFrame.Shape.NoFrame)

        if sys.platform.lower() != "darwin":
            self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)



# -*- coding: utf-8 -*-

import os

# USER_HOME variable is design for mobile application (feeluownx)
USER_HOME = os.environ.get('FEELUOWN_USER_HOME', os.path.expanduser('~'))
HOME_DIR = os.path.expanduser(USER_HOME + '/.FeelUOwn')

DATA_DIR = HOME_DIR + '/data'
USER_PLUGINS_DIR = HOME_DIR + '/plugins'
USER_THEMES_DIR = HOME_DIR + '/themes'
CACHE_DIR = HOME_DIR + '/cache'
SONG_DIR = HOME_DIR + '/songs'
COLLECTIONS_DIR = HOME_DIR + '/collections'

LOG_FILE = HOME_DIR + '/stdout.log'
STATE_FILE = os.path.join(DATA_DIR, 'state.json')
DEFAULT_RCFILE_PATH = os.path.expanduser(f'{USER_HOME}/.fuorc')



from __future__ import annotations
import logging
import sys
from typing import TYPE_CHECKING

from PyQt6.QtCore import Qt, QSize
from PyQt6.QtGui import QGuiApplication, QResizeEvent
from PyQt6.QtWidgets import (
    QWidget,
    QHBoxLayout,
    QLabel,
    QVBoxLayout,
    QSizePolicy,
    QScrollArea,
    QFrame,
)

from feeluown.i18n import t
from feeluown.excs import ResourceNotFound
from feeluown.library import (
    SupportsSongHotComments,
    SupportsSongSimilar,
    ModelFlags,
)
from feeluown.player import Lyric
from feeluown.library import reverse, resolve
from feeluown.utils import aio
from feeluown.utils.aio import run_afn
from feeluown.utils.reader import create_reader
from feeluown.gui.components import SongMVTextButton
from feeluown.gui.helpers import BgTransparentMixin, fetch_cover_wrapper, resize_font
from feeluown.gui.widgets.labels import ElidedLineLabel
from feeluown.gui.widgets.header import LargeHeader, MidHeader
from feeluown.gui.widgets.textbtn import TextButton
from feeluown.gui.widgets.cover_label import CoverLabelV2
from feeluown.gui.widgets.comment_list import CommentListView, CommentListModel
from feeluown.gui.widgets.lyric import LyricView
from feeluown.gui.widgets.song_minicard_list import (
    SongMiniCardListDelegate,
    SongMiniCardListModel,
    SongMiniCardListView,
)
from .template import render_error_message

if TYPE_CHECKING:
    from feeluown.app.gui_app import GuiApp

logger = logging.getLogger(__name__)


def or_unknown(x):
    return x or t("unknown")


async def render(req, **kwargs):  # pylint: disable=too-many-locals,too-many-branches
    app: GuiApp = req.ctx["app"]
    song = req.ctx["model"]

    provider = app.library.get(song.source)
    if provider is None:
        await render_error_message(
            app, t("track-source-provider-missing", providerName=song.source)
        )
        return

    # TODO: Initialize the view with song object, and it should reduce
    # the code complexity.
    view = SongExploreView(app=app)
    app.ui.right_panel.set_body(view)

    try:
        upgraded_song = await aio.run_fn(app.library.song_upgrade, song)
    except ResourceNotFound:
        upgraded_song = None
    if upgraded_song is not None:
        song = upgraded_song

    # Before fetching more data, show some widgets.
    view.title_label.set_src_text(song.title)
    view.title_label.setToolTip(song.title)
    view.play_btn.clicked.connect(lambda: app.playlist.play_model(song))

    # Show other widgets.
    await view.maybe_show_web_url_btn(provider, song)

    # Fetch album.
    album = upgraded_song.album if upgraded_song is not None else None
    if album is not None:
        try:
            upgraded_album = await aio.run_fn(app.library.album_upgrade, song.album)
        except ResourceNotFound:
            upgraded_album = None
    else:
        upgraded_album = None
    if upgraded_album is not None:
        album = upgraded_album

    run_afn(view.show_song_wiki, song, album)
    run_afn(view.maybe_show_song_pic, upgraded_song, upgraded_album)
    run_afn(view.maybe_show_song_similar, provider, song)
    run_afn(view.maybe_show_mv_btn, song)
    run_afn(view.maybe_show_song_lyric, song)
    try:
        await view.maybe_show_song_hot_comments(provider, song)
    except:  # noqa
        logger.exception("show song hot comments failed")


class ScrollArea(QScrollArea, BgTransparentMixin):
    def __init__(self, app: GuiApp, parent=None):
        super().__init__(parent=parent)

        self._app = app

        self.setWidgetResizable(True)
        self.setFrameShape(QFrame.Shape.NoFrame)

        # As far as I know, KDE and GNOME can't auto hide the scrollbar,
        # and they show an old-fation vertical scrollbar.
        # HELP: implement an auto-hide scrollbar for Linux
        if sys.platform.lower() != "darwin":
            self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)


class HeaderLabel(QLabel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setTextFormat(Qt.TextFormat.RichText)
        self.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse)


class LyricLabel(QLabel, BgTransparentMixin):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setWordWrap(True)
        self.setTextInteractionFlags(Qt.TextInteractionFlag.TextSelectableByMouse)
        self.setAlignment(Qt.AlignmentFlag.AlignHCenter)
        font = self.font()
        resize_font(font, 0)
        self.setFont(font)
        self.setTextFormat(Qt.TextFormat.RichText)


class SongWikiLabel(QLabel):
    def __init__(self, app: GuiApp, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._app = app
        self.setTextFormat(Qt.TextFormat.RichText)
        self.setTextInteractionFlags(Qt.TextInteractionFlag.TextBrowserInteraction)
        self.linkActivated.connect(self.on_link_activated)
        self.setWordWrap(True)

    async def show_song(self, song, album):
        artists_str_list = []
        for artist in song.artists:
            artists_str_list.append(
                f'<a href="{reverse(artist)}">{artist.name_display}</a>'
            )
        artists_str = " / ".join(artists_str_list)
        # Only show the date, like yyyy-mm-dd. Do not show hour/minutes/seconds.
        if song.date:
            date = song.date
        elif album is not None and ModelFlags.normal in ModelFlags(album.meta.flags):
            date = album.released
        else:
            date = ""
        date_fmted = date[:10] if date else ""
        album_str = (
            f'<a href="{reverse(album)}">{album.name_display}</a>' if album else ""
        )
        kvs = [
            (t("musician"), artists_str),
            (t("track-belongs-album"), or_unknown(album_str)),
            (t("release-date"), or_unknown(date_fmted)),
            (t("track-genre"), or_unknown(song.genre)),
        ]
        lines = []
        for k, v in kvs:
            line = f"<span style='color: grey'>{k}：</span>" + v
            lines.append(line)
        self.setText("<br/>".join(lines))

    def on_link_activated(self, url):
        model = resolve(url)
        self._app.browser.goto(model=model)


class LeftCon(QWidget, BgTransparentMixin):
    def sizeHint(self):
        size_hint = super().sizeHint()
        return QSize(600, size_hint.height())


class RightCon(QWidget):
    def sizeHint(self):
        size_hint = super().sizeHint()
        return QSize(130, size_hint.height())


class Title(LargeHeader, ElidedLineLabel):
    pass


class SongExploreView(QWidget):
    def __init__(self, app: GuiApp):
        super().__init__(parent=None)
        self._app = app

        self._title_cover_spacing = 10
        self._left_right_spacing = 30
        self.title_label = Title()
        self.similar_songs_header = MidHeader(t("similar-tracks"))
        self.comments_header = MidHeader(t("track-hot-comments"))
        self.lyric_view = LyricView(parent=self)
        self.play_btn = TextButton(t("track-start-play"))
        self.play_mv_btn = SongMVTextButton(self._app)
        self.copy_web_url_btn = TextButton(t("track-webpage-url-copy"))
        self.cover_label = CoverLabelV2(app=app)
        self.song_wiki_label = SongWikiLabel(app)
        self.comments_view = CommentListView()
        self.similar_songs_view = SongMiniCardListView(row_height=60, no_scroll_v=True)
        delegate = SongMiniCardListDelegate(
            self.similar_songs_view,
            card_min_width=150,
            card_height=40,
            card_padding=(5 + SongMiniCardListDelegate.img_padding, 5, 0, 5),
            card_right_spacing=10,
        )
        self.similar_songs_view.setItemDelegate(delegate)

        self._setup_ui()

        self.similar_songs_view.play_song_needed.connect(app.playlist.play_model)

    def _setup_ui(self):
        self.lyric_view.setSizePolicy(
            QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Expanding
        )
        self.cover_label.setFixedSize(160, 160)
        self.cover_label.setSizePolicy(
            QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Preferred
        )

        self._left_con = LeftCon(self)  #: left container
        self._left_con_scrollarea = ScrollArea(self._app)
        self._left_con_scrollarea.setWidget(self._left_con)
        self._right_con = RightCon(self)
        self._left_layout = QVBoxLayout(self._left_con)
        self._right_layout = QVBoxLayout(self._right_con)
        self._layout = QHBoxLayout(self)

        self._left_top_layout = QHBoxLayout()
        self._song_meta_layout = QVBoxLayout()
        self._btns_layout = QHBoxLayout()
        self._btns_layout.setSpacing(6)

        self._layout.setSpacing(0)
        self._layout.setContentsMargins(10, 20, 10, 0)

        self._layout.addWidget(self._left_con_scrollarea, 3)
        self._layout.addSpacing(self._left_right_spacing)
        self._layout.addWidget(self._right_con, 1)

        self._left_layout.addLayout(self._left_top_layout)
        self._left_layout.addSpacing(self._title_cover_spacing)
        self._left_layout.addWidget(self.comments_header)
        self._left_layout.addWidget(self.comments_view)
        self._left_layout.addWidget(self.similar_songs_header)
        self._left_layout.addWidget(self.similar_songs_view)
        self._left_layout.addSpacing(10)

        self._left_top_layout.addWidget(
            self.cover_label, alignment=Qt.AlignmentFlag.AlignTop
        )
        self._left_top_layout.addSpacing(10)
        self._left_top_layout.addLayout(self._song_meta_layout)

        self._song_meta_layout.addWidget(self.title_label)
        self._song_meta_layout.addStretch(0)
        self._song_meta_layout.addWidget(
            self.song_wiki_label, alignment=Qt.AlignmentFlag.AlignTop
        )
        self._song_meta_layout.addStretch(0)
        self._song_meta_layout.addLayout(self._btns_layout)

        self._btns_layout.addWidget(self.play_btn)
        self._btns_layout.addWidget(self.play_mv_btn)
        self._btns_layout.addWidget(self.copy_web_url_btn)
        self._btns_layout.addStretch(0)

        self._right_layout.addWidget(self.lyric_view)

    async def maybe_show_web_url_btn(self, provider, song):
        try:
            web_url = self._app.library.song_get_web_url(song)
        except ResourceNotFound:
            self.copy_web_url_btn.hide()
        else:
            # TODO: Open url in browser when alt key is pressed. Use
            # QDesktopServices.openUrl to open url in browser, and
            # you may use QGuiApplication::keyboardModifiers to check
            # if alt key is pressed.
            #
            # NOTE(cosven): Since switching from applications is inconvenience,
            # the default behaviour of button is url-copy instead of url-open.

            async def copy_song_web_url():
                QGuiApplication.clipboard().setText(web_url)
                self._app.show_msg(t("track-webpage-url-copied", url=web_url))

            self.copy_web_url_btn.clicked.connect(
                lambda: aio.run_afn(copy_song_web_url)
            )

    async def show_song_wiki(self, song, album):
        aio.run_afn(self.song_wiki_label.show_song, song, album)

    async def maybe_show_song_similar(self, provider, song):
        if isinstance(provider, SupportsSongSimilar):
            songs = await aio.run_fn(provider.song_list_similar, song)
            model = SongMiniCardListModel(
                create_reader(songs), fetch_cover_wrapper(self._app)
            )
            self.similar_songs_view.setModel(model)
        else:
            msg = t(
                "error-message-template",
                interface=SupportsSongSimilar.__name__,
            ).format(feature=t("find-similar-tracks"))

            self.similar_songs_header.setText(msg)

    async def maybe_show_song_hot_comments(self, provider, song):
        if isinstance(provider, SupportsSongHotComments):
            comments = await aio.run_fn(provider.song_list_hot_comments, song)
            comments_reader = create_reader(comments)
            self.comments_view.setModel(CommentListModel(comments_reader))
        else:
            msg = t(
                "error-message-template",
                interface=SupportsSongHotComments.__name__,
            ).format(feature=t("track-view-comments"))

            self.similar_songs_header.setText(msg)
            self.comments_header.setText(msg)

    async def maybe_show_mv_btn(self, song):
        self.play_mv_btn.bind_song(song)
        await self.play_mv_btn.get_mv()

    async def maybe_show_song_lyric(self, song):
        if self._app.playlist.current_song == song:
            lyric = self._app.live_lyric.current_lyrics[0]
            self.lyric_view.set_lyric(lyric)
            self._app.live_lyric.line_changed.connect(
                self.lyric_view.on_line_changed, weak=True
            )
            return

        lyric_model = self._app.library.song_get_lyric(song)
        if lyric_model is None:
            return
        self.lyric_view.set_lyric(Lyric.from_content(lyric_model.content))

    async def maybe_show_song_pic(self, song, album):
        if album:
            cover_media = await aio.run_fn(
                self._app.library.model_get_cover_media, album
            )
            aio.run_afn(
                self.cover_label.show_cover_media,
                cover_media,
                reverse(album) + "/cover",
            )
        else:
            cover_media = await aio.run_fn(
                self._app.library.model_get_cover_media, song
            )
            aio.run_afn(
                self.cover_label.show_cover_media,
                cover_media,
                reverse(song) + "/pic_url",
            )

    def resizeEvent(self, e: QResizeEvent) -> None:
        margins = self.layout().contentsMargins()
        margin_h = margins.left() + margins.right()
        self._left_con.setMaximumWidth(
            self.width() - margin_h - self._left_right_spacing - self._right_con.width()
        )
        return super().resizeEvent(e)



import asyncio
import logging
import time
from typing import List

from PyQt6.QtCore import Qt, QSize, QRect
from PyQt6.QtGui import QPalette, QResizeEvent, QFontMetrics
from PyQt6.QtWidgets import (
    QFrame,
    QHBoxLayout,
    QLabel,
    QListWidget,
    QListWidgetItem,
    QVBoxLayout,
    QWidget,
)

from feeluown.ai import AISongModel, AISongMatcher
from feeluown.app.gui_app import GuiApp
from feeluown.gui.consts import ScrollBarWidth
from feeluown.gui.widgets import PlayButton, PlusButton
from feeluown.gui.widgets.header import MidHeader
from feeluown.gui.widgets.ai_chat import ChatHistoryWidget, ChatInputWidget
from feeluown.gui.components.overlay import AppOverlayContainer
from feeluown.i18n import t
from feeluown.utils import aio

logger = logging.getLogger(__name__)


class AISongItemWidget(QWidget):
    """Display a single AISongModel with play support."""

    def __init__(self, app: "GuiApp", ai_song: AISongModel, parent=None):
        super().__init__(parent=parent)
        self._app = app

        self._ai_song = ai_song
        self._matched_song = None
        self._match_lock = asyncio.Lock()

        self._title_label = QLabel(f"{ai_song.title} • {ai_song.artists_name}")
        self._desc_label = QLabel(ai_song.description)
        self._desc_label.setWordWrap(True)
        self._play_btn = PlayButton(length=16, padding=0.22)
        self._add_to_playlist_btn = PlusButton(length=16, padding=0.25)
        self._btns = [self._play_btn, self._add_to_playlist_btn]

        self._play_btn.clicked.connect(self._on_play_clicked)
        self._add_to_playlist_btn.clicked.connect(self._on_add_to_playlist_clicked)

        self._setup_ui()

    def _setup_ui(self):
        font = self._desc_label.font()
        font.setPixelSize(11)
        self._desc_label.setFont(font)
        palette = self._desc_label.palette()
        palette.setColor(
            QPalette.ColorRole.Text,
            palette.color(QPalette.ColorGroup.Disabled, QPalette.ColorRole.Text),
        )
        self._desc_label.setPalette(palette)

        layout = QHBoxLayout(self)
        layout.setContentsMargins(10, 5, ScrollBarWidth + 5, 5)
        layout.setSpacing(5)

        self.label_layout = label_layout = QVBoxLayout()
        label_layout.addWidget(self._title_label)
        label_layout.addWidget(self._desc_label)
        label_layout.addStretch(0)

        self.btn_layout = btn_layout = QHBoxLayout()
        btn_layout.setSpacing(0)
        for btn in self._btns:
            btn_layout.addWidget(btn)

        layout.addLayout(label_layout)
        layout.addLayout(btn_layout)

    def height_for_width(self, width) -> QSize:
        desc_fm = QFontMetrics(self._desc_label.font())
        h_spacing = (self.layout().spacing() +
                     self.btn_layout.spacing() * (len(self._btns) - 1))
        v_spacing = self.layout().spacing()
        margins = self.layout().contentsMargins()
        h_margins = margins.left() + margins.right()
        v_margins = margins.top() + margins.bottom()
        btn_width = self._play_btn.width() + self._add_to_playlist_btn.width()
        desc_width = width - btn_width - h_spacing - h_margins
        desc_height = desc_fm.boundingRect(
            QRect(0, 0, desc_width, 0), Qt.TextFlag.TextWordWrap, self._desc_label.text()
        ).height()
        title_height = 16
        total_height = title_height + desc_height + v_spacing + v_margins
        return total_height

    def _on_play_clicked(self):
        if self._matched_song is not None:
            self._app.playlist.play_model(self._matched_song)
        else:
            aio.run_afn_ref(self.match_and_play)

    def _on_add_to_playlist_clicked(self):
        if self._matched_song is not None:
            self._app.playlist.add(self._matched_song)
        else:
            aio.run_afn_ref(self.match_and_add_to_playlist)

    async def match_and_play(self):
        await self.match()
        if self._matched_song:
            self._app.playlist.play_model(self._matched_song)

    async def match_and_add_to_playlist(self):
        await self.match()
        if self._matched_song:
            self._app.playlist.add(self._matched_song)

    async def match(self):
        async with self._match_lock:
            return await self.match_no_lock()

    async def match_no_lock(self):
        if self._matched_song is not None:
            return self._matched_song

        for btn in self._btns:
            btn.setEnabled(False)
            btn.setToolTip(t("ai-chat-match-resource"))
        try:
            self._matched_song = await AISongMatcher(self._app).match(self._ai_song)
        except Exception:
            logger.exception("match ai song failed")
            for btn in self._btns:
                btn.setEnabled(True)
        else:
            if self._matched_song is not None:
                for btn in self._btns:
                    btn.setEnabled(True)
            else:
                for btn in self._btns:
                    btn.setToolTip(t("ai-chat-match-resource-failed"))


class AISongListWidget(QWidget):
    """Container widget to display AISongModel list."""

    def __init__(self, app: "GuiApp", parent=None):
        super().__init__(parent=parent)
        self._app = app

        self._title_label = QLabel(t("ai-chat-track-candidate-list"))
        self._title_label.setAlignment(Qt.AlignmentFlag.AlignCenter)
        self._list_widget = QListWidget(self)
        self._list_widget.setFrameShape(QFrame.Shape.NoFrame)
        self._list_widget.setSelectionMode(QListWidget.SelectionMode.NoSelection)
        self._list_widget.setResizeMode(QListWidget.ResizeMode.Adjust)
        self._list_widget.setAlternatingRowColors(True)

        layout = QVBoxLayout(self)
        layout.setContentsMargins(5, 5, 5, 5)
        layout.setSpacing(10)
        layout.addWidget(self._title_label)
        layout.addWidget(self._list_widget)

        self.setAutoFillBackground(True)

    def set_ai_songs(self, ai_songs: List[AISongModel]):
        self._clear_items()
        for ai_song in ai_songs:
            item = AISongItemWidget(self._app, ai_song, parent=self)
            list_item = QListWidgetItem(self._list_widget)
            width = self._get_width_for_items()
            list_item.setSizeHint(QSize(width, item.height_for_width(width)))
            self._list_widget.addItem(list_item)
            self._list_widget.setItemWidget(list_item, item)

    def _clear_items(self):
        self._list_widget.clear()

    def _get_width_for_items(self):
        margins = self.layout().contentsMargins()
        return self.width() - ScrollBarWidth * 2 - margins.left() - margins.right()

    def resizeEvent(self, event: QResizeEvent):
        """Update item sizes when widget is resized."""
        super().resizeEvent(event)
        self._update_item_sizes()

    def _update_item_sizes(self):
        """Update all item sizes based on current viewport width."""
        width = self._get_width_for_items()
        if width <= 0:
            return

        for i in range(self._list_widget.count()):
            list_item = self._list_widget.item(i)
            item_widget = self._list_widget.itemWidget(list_item)
            if item_widget is not None:
                height = item_widget.height_for_width(width)
                list_item.setSizeHint(QSize(width, height))


class AIChatBox(QWidget):
    """
    A lightweight AI chat UI for the AI Radio page.
    This only provides UI; real message sending is intentionally left empty.
    """

    def __init__(self, app, parent=None):
        super().__init__(parent=parent)
        self._app = app
        self.copilot = self._app.ai.get_copilot()

        self._header = MidHeader(t("ai-chat-header"))
        self._new_thread_btn = PlusButton(length=12)
        self._new_thread_btn.setToolTip(t("ai-chat-new"))
        self.history_widget = ChatHistoryWidget(self)
        self.input_widget = ChatInputWidget(self)

        self.input_widget.send_clicked.connect(
            lambda q: aio.run_afn_ref(self.exec_user_query, q)
        )
        self._new_thread_btn.clicked.connect(self.on_new_thread_btn_clicked)
        self.copilot.working_state_changed.connect(self.on_working_state_changed)
        self.setup_ui()

    def setup_ui(self):
        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(8)
        header_layout = QHBoxLayout()
        header_layout.addWidget(self._header)
        header_layout.addWidget(self._new_thread_btn)
        header_layout.addStretch(0)
        layout.addLayout(header_layout)
        layout.addWidget(self.history_widget)
        layout.addWidget(self.input_widget)

    async def exec_user_query(self, query: str):
        self.history_widget.add_message("user", query)
        self.history_widget.scroll_to_bottom()

        current_label = None
        response_message = ""
        last_update_ts = time.time()

        async for token, metadata in self.copilot.astream_user_query(query):
            node = metadata["langgraph_node"]
            if node == "model":
                if current_label is None:
                    current_label = self.history_widget.create_message_label(
                        "assistant", ""
                    )
                for block in token.content_blocks:
                    if block["type"] == "text":
                        response_message += block["text"]
                if last_update_ts + 0.2 <= time.time():
                    current_label.setText(response_message)
                    self.history_widget.scroll_to_bottom()
                    last_update_ts = time.time()
            elif node == "tools":
                self.history_widget.create_message_label(
                    "tools", f"tools: {token.name}"
                )
                current_label = None

        if current_label is not None:
            current_label.setText(response_message)
        self.history_widget.scroll_to_bottom()

    def on_new_thread_btn_clicked(self):
        self.copilot.new_thread()
        self.history_widget.clear()

    def on_working_state_changed(self, working: bool):
        self.input_widget.enable_send(not working)


class Body(QWidget):
    def __init__(self, app: "GuiApp"):
        super().__init__(parent=None)
        self._app = app
        copilot = self._app.ai.get_copilot()
        copilot.candidates_changed.connect(
            self.on_copilot_candidates_changed, aioqueue=True)

        self._list_widget = AISongListWidget(app, self)
        self._chat_box = AIChatBox(app)

        self._layout = QHBoxLayout(self)
        self._layout.setContentsMargins(10, 10, 10, 10)
        self._layout.setSpacing(10)
        self._layout.addWidget(self._chat_box)
        self._layout.addWidget(self._list_widget)
        self._layout.setStretch(0, 2)
        self._layout.setStretch(1, 1)
        self.setAutoFillBackground(True)

    def on_copilot_candidates_changed(self, candidates):
        self._list_widget.set_ai_songs(candidates)


def create_aichat_overlay(app: "GuiApp", parent=None) -> AppOverlayContainer:
    """Create an overlay for the AI chat"""
    body = Body(app)
    overlay = AppOverlayContainer(app, body, parent=parent)
    return overlay


if __name__ == "__main__":
    from feeluown.gui.debug import mock_app, simple_layout

    with simple_layout(theme="dark") as layout, mock_app() as app:
        app.size.return_value = QSize(600, 400)
        overlay = create_aichat_overlay(app)
        long_description = (
            "Very long description"
            "Ahhhhhhhhhhhhhhhhhhhhhhhhhhhhhh"
            "Ahhhhhhhhhhhhhhhhhhhhhhhhhhhhhh"
        )
        sample_ai_songs = [
            AISongModel(
                title="Song A", artists_name="Artist A", description=long_description
            ),
            AISongModel(title="Song B", artists_name="Artist B", description="Desc B"),
            AISongModel(title="Song C", artists_name="Artist C", description="Desc C"),
        ]

        overlay.body._chat_box.history_widget.add_message("user", "hello world")
        overlay.body._chat_box.history_widget.add_message("assistant", "Hi, 我是你的音乐助手")
        overlay.body._chat_box.history_widget.add_message("tools", "tools: play_model")
        overlay.body._list_widget.set_ai_songs(sample_ai_songs)
        overlay.show()
        layout.addWidget(overlay)



import sys
from contextlib import contextmanager

from PyQt6.QtCore import QMetaObject, pyqtSlot, QSize
from PyQt6.QtGui import QOpenGLContext

from feeluown.mpv import (
    MpvRenderContext,
    OpenGlCbGetProcAddrFn,
    _mpv_set_property_string,
)

from feeluown.gui.widgets.video import VideoOpenGLWidget


def get_proc_addr(_, name: bytes):
    glctx = QOpenGLContext.currentContext()
    if glctx is None:
        return 0
    addr = int(glctx.getProcAddress(name))
    return addr


class MpvOpenGLWidget(VideoOpenGLWidget):
    """Mpv video output window

    On destruction, the shutdown method should be called to release resources.

    This Widget is modeled after some C++ programs
    (see examples in the project’s research directory),
    we don’t fully understand the underlying principles, but testing shows it works fine.

    The main current questions:

    - [ ] The shutdown method is expected to release resources;
          is the current implementation reasonable?
          When should the shutdown method be called?
    """

    def __init__(self, app, parent=None):
        super().__init__(app=app, parent=parent)
        self._app = app
        self.mpv = self._app.player._mpv  # noqa
        self.ctx = None
        self.get_proc_addr_c = OpenGlCbGetProcAddrFn(get_proc_addr)

    def initializeGL(self):
        params = {"get_proc_address": self.get_proc_addr_c}
        self.ctx = MpvRenderContext(self.mpv, "opengl", opengl_init_params=params)
        self.ctx.update_cb = self.on_update

    def shutdown(self):
        if self.ctx is not None:
            self.ctx.free()
            self.ctx = None

    def paintGL(self):
        # HELP: It seems that `initializeGL` is called by Qt on
        # old version (<= v5.15.2)
        if self.ctx is None:
            self.initializeGL()
            assert self.ctx is not None
        # compatible with HiDPI display
        ratio = self._app.devicePixelRatio()
        w = int(self.width() * ratio)
        h = int(self.height() * ratio)
        opengl_fbo = {"w": w, "h": h, "fbo": self.defaultFramebufferObject()}
        self.ctx.render(flip_y=True, opengl_fbo=opengl_fbo)

        for gl_painter in self._gl_painters:
            gl_painter.paint(self)

    @pyqtSlot()
    def maybe_update(self):
        if self.window().isMinimized():
            self.makeCurrent()
            self.paintGL()
            self.context().swapBuffers(self.context().surface())
            self.doneCurrent()
        else:
            self.update()

    def on_update(self, ctx=None):
        # maybe_update method should run on the thread that creates the
        # OpenGLContext, which in general is the main thread.
        # QMetaObject.invokeMethod can do this trick.
        QMetaObject.invokeMethod(self, "maybe_update")

    # NOTE(cosven): heightForwidth does not work (tested inside nowplaying_overlay.py)
    def hasHeightForWidth(self) -> bool:
        return bool(self.mpv.width)

    def heightForWidth(self, width: int) -> int:
        if self.mpv.width:
            return width // self.mpv.width * self.mpv.height
        return super().heightForWidth(width)

    def sizeHint(self):
        if self.mpv.width:
            return QSize(self.mpv.width, self.mpv.height)
        return super().sizeHint()

    @contextmanager
    def change_parent(self):
        # on macOS (with PyQt5), changing mpv widget parent cause no side effects.
        #
        # on Linux (wayland), it seems changing mpv widget parent may cause segfault,
        # so do some hack to avoid crash.
        #
        # on macOS (with PyQt6), changing mpv widget parent also causes
        # 'RuntimeError: Unspecified error'
        self._before_change_mpv_widget_parent()
        try:
            yield
        finally:
            self._after_change_mpv_widget_parent()

    def _before_change_mpv_widget_parent(self):
        """
        According to Qt docs, reparenting an OpenGLWidget will destory the GL context.
        In mpv widget, it calls _mpv_opengl_cb_uninit_gl. After uninit_gl, mpv can't show
        video anymore because video_out is destroyed.

        See mpv mpv_opengl_cb_uninit_gl implementation for more details.
        """
        _mpv_set_property_string(self.mpv.handle, b"vid", b"no")
        self.hide()
        self.shutdown()

    def _after_change_mpv_widget_parent(self):
        """
        To recover the video show, we should reinit gl and reinit video. gl is
        automatically reinited when the mpv_widget do painting. We should
        manually reinit video.

        NOTE(cosven): After some investigation, I found that the API in mpv to
        reinit video_out(maybe video is almost same as video_out)
        is init_best_video_out. Theoretically, sending 'video-reload' command
        will trigger this API. However, we can't run this command
        in our case and I don't know why. Another way to trigger
        the API is to switch track. Changing vid property just switch the track.

        Inpect mpv init_best_video_out caller for more details. You should see
        mp_switch_track_n is one of the entrypoint.

        Note the GL must be re-initialized properly before this function is called.
        Generally, show parent once can trigger the initialization. Otherwise,
        mpv.vid will not be set to 1, and video can not be shown.
        """
        self.show()
        self.repaint()  # force repaint to trigger re-initialization
        _mpv_set_property_string(self.mpv.handle, b"vid", b"1")
        if not bool(self.mpv.vid):
            print("WARNING: video widget is not reconfigured properly", file=sys.stderr)


# TODO: Implement MpvEmbeddedWidget
class MpvEmbeddedWidget:
    pass
